"""
Context
-------
A Context is a data structure that is used to store information about the current state of a conversation.
It is used to keep track of the user's input, the current stage of the conversation, and any other
information that is relevant to the current context of a dialog.
The Context provides a convenient interface for working with data, allowing developers to easily add,
retrieve, and manipulate data as the conversation progresses.

The Context data structure provides several key features to make working with data easier.
Developers can use the context to store any information that is relevant to the current conversation,
such as user data, session data, conversation history, or etc.
This allows developers to easily access and use this data throughout the conversation flow.

Another important feature of the context is data serialization.
The context can be easily serialized to a format that can be stored or transmitted, such as JSON.
This allows developers to save the context data and resume the conversation later.
"""

from __future__ import annotations
import logging
from uuid import UUID, uuid4
from typing import Any, Optional, Union, Dict, List, Set, TYPE_CHECKING

from pydantic import BaseModel, Field, field_validator

from chatsky.script.core.message import Message
from chatsky.script.core.types import NodeLabel2Type
from chatsky.pipeline.types import ComponentExecutionState
from chatsky.slots.slots import SlotManager

if TYPE_CHECKING:
    from chatsky.script.core.script import Node

logger = logging.getLogger(__name__)


def get_last_index(dictionary: dict) -> int:
    """
    Obtain the last index from the `dictionary`. Return `-1` if the `dict` is empty.

    :param dictionary: Dictionary with unsorted keys.
    :return: Last index from the `dictionary`.
    """
    indices = list(dictionary)
    return indices[-1] if indices else -1


class FrameworkData(BaseModel):
    """
    Framework uses this to store data related to any of its modules.
    """

    service_states: Dict[str, ComponentExecutionState] = Field(default_factory=dict, exclude=True)
    "Statuses of all the pipeline services. Cleared at the end of every turn."
    actor_data: Dict[str, Any] = Field(default_factory=dict, exclude=True)
    "Actor service data. Cleared at the end of every turn."
    stats: Dict[str, Any] = Field(default_factory=dict)
    "Enables complex stats collection across multiple turns."
    slot_manager: SlotManager = Field(default_factory=SlotManager)
    "Stores extracted slots."


class Context(BaseModel):
    """
    A structure that is used to store data about the context of a dialog.

    Avoid storing unserializable data in the fields of this class in order for
    context storages to work.
    """

    id: Union[UUID, int, str] = Field(default_factory=uuid4)
    """
    `id` is the unique context identifier. By default, randomly generated using `uuid4` `id` is used.
    `id` can be used to trace the user behavior, e.g while collecting the statistical data.
    """
    labels: Dict[int, NodeLabel2Type] = Field(default_factory=dict)
    """
    `labels` stores the history of all passed `labels`

        - key - `id` of the turn.
        - value - `label` on this turn.
    """
    requests: Dict[int, Message] = Field(default_factory=dict)
    """
    `requests` stores the history of all `requests` received by the agent

        - key - `id` of the turn.
        - value - `request` on this turn.
    """
    responses: Dict[int, Message] = Field(default_factory=dict)
    """
    `responses` stores the history of all agent `responses`

        - key - `id` of the turn.
        - value - `response` on this turn.
    """
    misc: Dict[str, Any] = Field(default_factory=dict)
    """
    `misc` stores any custom data. The scripting doesn't use this dictionary by default,
    so storage of any data won't reflect on the work on the internal Chatsky Scripting functions.

    Avoid storing unserializable data in order for context storages to work.

        - key - Arbitrary data name.
        - value - Arbitrary data.
    """
    framework_data: FrameworkData = Field(default_factory=FrameworkData)
    """
    This attribute is used for storing custom data required for pipeline execution.
    It is meant to be used by the framework only. Accessing it may result in pipeline breakage.
    """

    @field_validator("labels", "requests", "responses")
    @classmethod
    def sort_dict_keys(cls, dictionary: dict) -> dict:
        """
        Sort the keys in the `dictionary`. This needs to be done after deserialization,
        since the keys are deserialized in a random order.

        :param dictionary: Dictionary with unsorted keys.
        :return: Dictionary with sorted keys.
        """
        return {key: dictionary[key] for key in sorted(dictionary)}

    @classmethod
    def cast(cls, ctx: Optional[Union[Context, dict, str]] = None, *args, **kwargs) -> Context:
        """
        Transform different data types to the objects of the
        :py:class:`~.Context` class.
        Return an object of the :py:class:`~.Context`
        type that is initialized by the input data.

        :param ctx: Data that is used to initialize an object of the
            :py:class:`~.Context` type.
            An empty :py:class:`~.Context` object is returned if no data is given.
        :return: Object of the :py:class:`~.Context`
            type that is initialized by the input data.
        """
        if not ctx:
            ctx = Context(*args, **kwargs)
        elif isinstance(ctx, dict):
            ctx = Context.model_validate(ctx)
        elif isinstance(ctx, str):
            ctx = Context.model_validate_json(ctx)
        elif not isinstance(ctx, Context):
            raise ValueError(
                f"Context expected to be an instance of the Context class "
                f"or an instance of the dict/str(json) type. Got: {type(ctx)}"
            )
        return ctx

    def add_request(self, request: Message):
        """
        Add a new `request` to the context.
        The new `request` is added with the index of `last_index + 1`.

        :param request: `request` to be added to the context.
        """
        request_message = Message.model_validate(request)
        last_index = get_last_index(self.requests)
        self.requests[last_index + 1] = request_message

    def add_response(self, response: Message):
        """
        Add a new `response` to the context.
        The new `response` is added with the index of `last_index + 1`.

        :param response: `response` to be added to the context.
        """
        response_message = Message.model_validate(response)
        last_index = get_last_index(self.responses)
        self.responses[last_index + 1] = response_message

    def add_label(self, label: NodeLabel2Type):
        """
        Add a new :py:data:`~.NodeLabel2Type` to the context.
        The new `label` is added with the index of `last_index + 1`.

        :param label: `label` that we need to add to the context.
        """
        last_index = get_last_index(self.labels)
        self.labels[last_index + 1] = label

    def clear(
        self,
        hold_last_n_indices: int,
        field_names: Union[Set[str], List[str]] = {"requests", "responses", "labels"},
    ):
        """
        Delete all records from the `requests`/`responses`/`labels` except for
        the last `hold_last_n_indices` turns.
        If `field_names` contains `misc` field, `misc` field is fully cleared.

        :param hold_last_n_indices: Number of last turns to keep.
        :param field_names: Properties of :py:class:`~.Context` to clear.
            Defaults to {"requests", "responses", "labels"}
        """
        field_names = field_names if isinstance(field_names, set) else set(field_names)
        if "requests" in field_names:
            for index in list(self.requests)[:-hold_last_n_indices]:
                del self.requests[index]
        if "responses" in field_names:
            for index in list(self.responses)[:-hold_last_n_indices]:
                del self.responses[index]
        if "misc" in field_names:
            self.misc.clear()
        if "labels" in field_names:
            for index in list(self.labels)[:-hold_last_n_indices]:
                del self.labels[index]
        if "framework_data" in field_names:
            self.framework_data = FrameworkData()

    @property
    def last_label(self) -> Optional[NodeLabel2Type]:
        """
        Return the last :py:data:`~.NodeLabel2Type` of
        the :py:class:`~.Context`.
        Return `None` if `labels` is empty.

        Since `start_label` is not added to the `labels` field,
        empty `labels` usually indicates that the current node is the `start_node`.
        """
        last_index = get_last_index(self.labels)
        return self.labels.get(last_index)

    @property
    def last_response(self) -> Optional[Message]:
        """
        Return the last `response` of the current :py:class:`~.Context`.
        Return `None` if `responses` is empty.
        """
        last_index = get_last_index(self.responses)
        return self.responses.get(last_index)

    @last_response.setter
    def last_response(self, response: Optional[Message]):
        """
        Set the last `response` of the current :py:class:`~.Context`.
        Required for use with various response wrappers.
        """
        last_index = get_last_index(self.responses)
        self.responses[last_index] = Message() if response is None else Message.model_validate(response)

    @property
    def last_request(self) -> Optional[Message]:
        """
        Return the last `request` of the current :py:class:`~.Context`.
        Return `None` if `requests` is empty.
        """
        last_index = get_last_index(self.requests)
        return self.requests.get(last_index)

    @last_request.setter
    def last_request(self, request: Optional[Message]):
        """
        Set the last `request` of the current :py:class:`~.Context`.
        Required for use with various request wrappers.
        """
        last_index = get_last_index(self.requests)
        self.requests[last_index] = Message() if request is None else Message.model_validate(request)

    @property
    def current_node(self) -> Optional[Node]:
        """
        Return current :py:class:`~chatsky.script.core.script.Node`.
        """
        actor_data = self.framework_data.actor_data
        node = (
            actor_data.get("processed_node")
            or actor_data.get("pre_response_processed_node")
            or actor_data.get("next_node")
            or actor_data.get("pre_transitions_processed_node")
            or actor_data.get("previous_node")
        )
        if node is None:
            logger.warning(
                "The `current_node` method should be called "
                "when an actor is running between the "
                "`ActorStage.GET_PREVIOUS_NODE` and `ActorStage.FINISH_TURN` stages."
            )

        return node



"""
Types
-----
The Types module contains several classes and special types that are used throughout `Chatsky Pipeline`.
The classes and special types in this module can include data models,
data structures, and other types that are defined for type hinting.
"""

from __future__ import annotations
from enum import unique, Enum
from typing import Callable, Union, Awaitable, Dict, List, Optional, Iterable, Any, Protocol, Hashable, TYPE_CHECKING
from typing_extensions import NotRequired, TypedDict, TypeAlias
from pydantic import BaseModel


if TYPE_CHECKING:
    from chatsky.pipeline.pipeline.pipeline import Pipeline
    from chatsky.pipeline.service.service import Service
    from chatsky.pipeline.service.group import ServiceGroup
    from chatsky.pipeline.service.extra import _ComponentExtraHandler
    from chatsky.messengers.common.interface import MessengerInterface
    from chatsky.context_storages import DBContextStorage
    from chatsky.script import Context, ActorStage, NodeLabel2Type, Script, Message


class PipelineRunnerFunction(Protocol):
    """
    Protocol for pipeline running.
    """

    def __call__(
        self, message: Message, ctx_id: Optional[Hashable] = None, update_ctx_misc: Optional[dict] = None
    ) -> Awaitable[Context]:
        """
        :param message: User request for pipeline to process.
        :param ctx_id:
            ID of the context that the new request belongs to.
            Optional, None by default.
            If set to `None`, a new context will be created with `message` being the first request.
        :param update_ctx_misc:
            Dictionary to be passed as an argument to `ctx.misc.update`.
            This argument can be used to store values in the `misc` dictionary before anything else runs.
            Optional; None by default.
            If set to `None`, `ctx.misc.update` will not be called.
        :return:
            Context instance that pipeline processed.
            The context instance has the id of `ctx_id`.
            If `ctx_id` is `None`, context instance has an id generated with `uuid.uuid4`.
        """


@unique
class ComponentExecutionState(str, Enum):
    """
    Enum, representing pipeline component execution state.
    These states are stored in `ctx.framework_keys.service_states`,
    that should always be requested with `NOT_RUN` being default fallback.
    Following states are supported:

    - NOT_RUN: component has not been executed yet (the default one),
    - RUNNING: component is currently being executed,
    - FINISHED: component executed successfully,
    - FAILED: component execution failed.
    """

    NOT_RUN = "NOT_RUN"
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    FAILED = "FAILED"


@unique
class GlobalExtraHandlerType(str, Enum):
    """
    Enum, representing types of global extra handlers, that can be set applied for a pipeline.
    The following types are supported:

    - BEFORE_ALL: function called before each pipeline call,
    - BEFORE: function called before each component,
    - AFTER: function called after each component,
    - AFTER_ALL: function called after each pipeline call.
    """

    BEFORE_ALL = "BEFORE_ALL"
    BEFORE = "BEFORE"
    AFTER = "AFTER"
    AFTER_ALL = "AFTER_ALL"


@unique
class ExtraHandlerType(str, Enum):
    """
    Enum, representing wrapper execution stage: before or after the wrapped function.
    The following types are supported:

    - UNDEFINED: extra handler function with undetermined execution stage,
    - BEFORE: extra handler function called before component,
    - AFTER: extra handler function called after component.
    """

    UNDEFINED = "UNDEFINED"
    BEFORE = "BEFORE"
    AFTER = "AFTER"


StartConditionCheckerFunction: TypeAlias = Callable[["Context", "Pipeline"], bool]
"""
A function type for components `start_conditions`.
Accepts context and pipeline, returns boolean (whether service can be launched).
"""


StartConditionCheckerAggregationFunction: TypeAlias = Callable[[Iterable[bool]], bool]
"""
A function type for creating aggregation `start_conditions` for components.
Accepts list of functions (other start_conditions to aggregate), returns boolean (whether service can be launched).
"""


ExtraHandlerConditionFunction: TypeAlias = Callable[[str], bool]
"""
A function type used during global extra handler initialization to determine
whether extra handler should be applied to component with given path or not.
Checks components path to be in whitelist (if defined) and not to be in blacklist (if defined).
Accepts str (component path), returns boolean (whether extra handler should be applied).
"""


class ServiceRuntimeInfo(BaseModel):
    """
    Type of object, that is passed to components in runtime.
    Contains current component info (`name`, `path`, `timeout`, `asynchronous`).
    Also contains `execution_state` - a dictionary,
    containing execution states of other components mapped to their paths.
    """

    name: str
    path: str
    timeout: Optional[float]
    asynchronous: bool
    execution_state: Dict[str, ComponentExecutionState]


ExtraHandlerFunction: TypeAlias = Union[
    Callable[["Context"], Any],
    Callable[["Context", "Pipeline"], Any],
    Callable[["Context", "Pipeline", "ExtraHandlerRuntimeInfo"], Any],
]
"""
A function type for creating extra handler (before and after functions).
Can accept current dialog context, pipeline, and current extra handler info.
"""


class ExtraHandlerRuntimeInfo(BaseModel):
    func: ExtraHandlerFunction
    stage: ExtraHandlerType
    component: ServiceRuntimeInfo


"""
Type of object, that is passed to wrappers in runtime.
Contains current wrapper info (`name`, `stage`).
Also contains `component` - runtime info of the component this wrapper is attached to.
"""


ServiceFunction: TypeAlias = Union[
    Callable[["Context"], None],
    Callable[["Context"], Awaitable[None]],
    Callable[["Context", "Pipeline"], None],
    Callable[["Context", "Pipeline"], Awaitable[None]],
    Callable[["Context", "Pipeline", ServiceRuntimeInfo], None],
    Callable[["Context", "Pipeline", ServiceRuntimeInfo], Awaitable[None]],
]
"""
A function type for creating service handlers.
Can accept current dialog context, pipeline, and current service info.
Can be both synchronous and asynchronous.
"""


ExtraHandlerBuilder: TypeAlias = Union[
    "_ComponentExtraHandler",
    TypedDict(
        "WrapperDict",
        {
            "timeout": NotRequired[Optional[float]],
            "asynchronous": NotRequired[bool],
            "functions": List[ExtraHandlerFunction],
        },
    ),
    List[ExtraHandlerFunction],
]
"""
A type, representing anything that can be transformed to ExtraHandlers.
It can be:

- ExtraHandlerFunction object
- Dictionary, containing keys `timeout`, `asynchronous`, `functions`
"""


ServiceBuilder: TypeAlias = Union[
    ServiceFunction,
    "Service",
    str,
    TypedDict(
        "ServiceDict",
        {
            "handler": "ServiceBuilder",
            "before_handler": NotRequired[Optional[ExtraHandlerBuilder]],
            "after_handler": NotRequired[Optional[ExtraHandlerBuilder]],
            "timeout": NotRequired[Optional[float]],
            "asynchronous": NotRequired[bool],
            "start_condition": NotRequired[StartConditionCheckerFunction],
            "name": Optional[str],
        },
    ),
]
"""
A type, representing anything that can be transformed to service.
It can be:

- ServiceFunction (will become handler)
- Service object (will be spread and recreated)
- String 'ACTOR' - the pipeline Actor will be placed there
- Dictionary, containing keys that are present in Service constructor parameters
"""


ServiceGroupBuilder: TypeAlias = Union[
    List[Union[ServiceBuilder, List[ServiceBuilder], "ServiceGroup"]],
    "ServiceGroup",
]
"""
A type, representing anything that can be transformed to service group.
It can be:

- List of `ServiceBuilders`, `ServiceGroup` objects and lists (recursive)
- `ServiceGroup` object (will be spread and recreated)
"""


PipelineBuilder: TypeAlias = TypedDict(
    "PipelineBuilder",
    {
        "messenger_interface": NotRequired[Optional["MessengerInterface"]],
        "context_storage": NotRequired[Optional[Union["DBContextStorage", Dict]]],
        "components": ServiceGroupBuilder,
        "before_handler": NotRequired[Optional[ExtraHandlerBuilder]],
        "after_handler": NotRequired[Optional[ExtraHandlerBuilder]],
        "optimization_warnings": NotRequired[bool],
        "parallelize_processing": NotRequired[bool],
        "script": Union["Script", Dict],
        "start_label": "NodeLabel2Type",
        "fallback_label": NotRequired[Optional["NodeLabel2Type"]],
        "label_priority": NotRequired[float],
        "condition_handler": NotRequired[Optional[Callable]],
        "handlers": NotRequired[Optional[Dict["ActorStage", List[Callable]]]],
    },
)
"""
A type, representing anything that can be transformed to pipeline.
It can be Dictionary, containing keys that are present in Pipeline constructor parameters.
"""



"""
Message Interfaces
------------------
The Message Interfaces module contains several basic classes that define the message interfaces.
These classes provide a way to define the structure of the messengers that are used to communicate with Chatsky.
"""

from __future__ import annotations
import abc
import asyncio
import logging
import uuid
from pathlib import Path
from tempfile import gettempdir
from typing import Optional, Any, List, Tuple, TextIO, Hashable, TYPE_CHECKING, Type

if TYPE_CHECKING:
    from chatsky.script import Context, Message
    from chatsky.pipeline.types import PipelineRunnerFunction
    from chatsky.messengers.common.types import PollingInterfaceLoopFunction
    from chatsky.script.core.message import Attachment

logger = logging.getLogger(__name__)


class MessengerInterface(abc.ABC):
    """
    Class that represents a message interface used for communication between pipeline and users.
    It is responsible for connection between user and pipeline, as well as for request-response transactions.
    """

    @abc.abstractmethod
    async def connect(self, pipeline_runner: PipelineRunnerFunction):
        """
        Method invoked when message interface is instantiated and connection is established.
        May be used for sending an introduction message or displaying general bot information.

        :param pipeline_runner: A function that should process user request and return context;
            usually it's a :py:meth:`~chatsky.pipeline.pipeline.pipeline.Pipeline._run_pipeline` function.
        """
        raise NotImplementedError


class MessengerInterfaceWithAttachments(MessengerInterface, abc.ABC):
    """
    MessengerInterface subclass that has methods for attachment handling.

    :param attachments_directory: Directory where attachments will be stored.
        If not specified, the temporary directory will be used.
    """

    supported_request_attachment_types: set[Type[Attachment]] = set()
    """
    Types of attachment that this messenger interface can receive.
    Attachments not in this list will be neglected.
    """

    supported_response_attachment_types: set[Type[Attachment]] = set()
    """
    Types of attachment that this messenger interface can send.
    Attachments not in this list will be neglected.
    """

    def __init__(self, attachments_directory: Optional[Path] = None) -> None:
        tempdir = gettempdir()
        if attachments_directory is not None and not str(attachments_directory.absolute()).startswith(tempdir):
            self.attachments_directory = attachments_directory
        else:
            warning_start = f"Attachments directory for {type(self).__name__} messenger interface"
            warning_end = "attachment data won't be cached locally!"
            if attachments_directory is None:
                self.attachments_directory = Path(tempdir) / f"chatsky-cache-{type(self).__name__}"
                logger.info(f"{warning_start} is None, so will be set to tempdir and {warning_end}")
            else:
                self.attachments_directory = attachments_directory
                logger.info(f"{warning_start} is in tempdir, so {warning_end}")
        self.attachments_directory.mkdir(parents=True, exist_ok=True)

    @abc.abstractmethod
    async def get_attachment_bytes(self, source: str) -> bytes:
        """
        Get attachment bytes from file source.

        E.g. if a file attachment consists of a URL of the file uploaded to the messenger servers,
        this method is the right place to call the messenger API for the file downloading.

        :param source: Identifying string for the file.
        :return: The attachment bytes.
        """
        raise NotImplementedError


class PollingMessengerInterface(MessengerInterface):
    """
    Polling message interface runs in a loop, constantly asking users for a new input.
    """

    @abc.abstractmethod
    def _request(self) -> List[Tuple[Message, Hashable]]:
        """
        Method used for sending users request for their input.

        :return: A list of tuples: user inputs and context ids (any user ids) associated with the inputs.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _respond(self, responses: List[Context]):
        """
        Method used for sending users responses for their last input.

        :param responses: A list of contexts, representing dialogs with the users;
            `last_response`, `id` and some dialog info can be extracted from there.
        """
        raise NotImplementedError

    def _on_exception(self, e: BaseException):
        """
        Method that is called on polling cycle exceptions, in some cases it should show users the exception.
        By default, it logs all exit exceptions to `info` log and all non-exit exceptions to `error`.

        :param e: The exception.
        """
        if isinstance(e, Exception):
            logger.error(f"Exception in {type(self).__name__} loop!", exc_info=e)
        else:
            logger.info(f"{type(self).__name__} has stopped polling.")

    async def _polling_loop(
        self,
        pipeline_runner: PipelineRunnerFunction,
        timeout: float = 0,
    ):
        """
        Method running the request - response cycle once.
        """
        user_updates = self._request()
        responses = [await pipeline_runner(request, ctx_id) for request, ctx_id in user_updates]
        self._respond(responses)
        await asyncio.sleep(timeout)

    async def connect(
        self,
        pipeline_runner: PipelineRunnerFunction,
        loop: PollingInterfaceLoopFunction = lambda: True,
        timeout: float = 0,
    ):
        """
        Method, running a request - response cycle in a loop.
        The looping behavior is determined by `loop` and `timeout`,
        for most cases the loop itself shouldn't be overridden.

        :param pipeline_runner: A function that should process user request and return context;
            usually it's a :py:meth:`~chatsky.pipeline.pipeline.pipeline.Pipeline._run_pipeline` function.
        :param loop: a function that determines whether polling should be continued;
            called in each cycle, should return `True` to continue polling or `False` to stop.
        :param timeout: a time interval between polls (in seconds).
        """
        while loop():
            try:
                await self._polling_loop(pipeline_runner, timeout)

            except BaseException as e:
                self._on_exception(e)
                break


class CallbackMessengerInterface(MessengerInterface):
    """
    Callback message interface is waiting for user input and answers once it gets one.
    """

    def __init__(self) -> None:
        self._pipeline_runner: Optional[PipelineRunnerFunction] = None

    async def connect(self, pipeline_runner: PipelineRunnerFunction):
        self._pipeline_runner = pipeline_runner

    async def on_request_async(
        self, request: Message, ctx_id: Optional[Hashable] = None, update_ctx_misc: Optional[dict] = None
    ) -> Context:
        """
        Method that should be invoked on user input.
        This method has the same signature as :py:class:`~chatsky.pipeline.types.PipelineRunnerFunction`.
        """
        return await self._pipeline_runner(request, ctx_id, update_ctx_misc)

    def on_request(
        self, request: Any, ctx_id: Optional[Hashable] = None, update_ctx_misc: Optional[dict] = None
    ) -> Context:
        """
        Method that should be invoked on user input.
        This method has the same signature as :py:class:`~chatsky.pipeline.types.PipelineRunnerFunction`.
        """
        return asyncio.run(self.on_request_async(request, ctx_id, update_ctx_misc))


class CLIMessengerInterface(PollingMessengerInterface):
    """
    Command line message interface is the default message interface, communicating with user via `STDIN/STDOUT`.
    This message interface can maintain dialog with one user at a time only.
    """

    def __init__(
        self,
        intro: Optional[str] = None,
        prompt_request: str = "request: ",
        prompt_response: str = "response: ",
        out_descriptor: Optional[TextIO] = None,
    ):
        super().__init__()
        self._ctx_id: Optional[Hashable] = None
        self._intro: Optional[str] = intro
        self._prompt_request: str = prompt_request
        self._prompt_response: str = prompt_response
        self._descriptor: Optional[TextIO] = out_descriptor

    def _request(self) -> List[Tuple[Message, Any]]:
        return [(Message(input(self._prompt_request)), self._ctx_id)]

    def _respond(self, responses: List[Context]):
        print(f"{self._prompt_response}{responses[0].last_response.text}; attachments: {responses[0].last_response.attachments}", file=self._descriptor)

    async def connect(self, pipeline_runner: PipelineRunnerFunction, **kwargs):
        """
        The CLIProvider generates new dialog id used to user identification on each `connect` call.

        :param pipeline_runner: A function that should process user request and return context;
            usually it's a :py:meth:`~chatsky.pipeline.pipeline.pipeline.Pipeline._run_pipeline` function.
        :param \\**kwargs: argument, added for compatibility with super class, it shouldn't be used normally.
        """
        self._ctx_id = uuid.uuid4()
        if self._intro is not None:
            print(self._intro)
        await super().connect(pipeline_runner, **kwargs)



"""
Pipeline
--------
The Pipeline module contains the :py:class:`.Pipeline` class,
which is a fundamental element of Chatsky. The Pipeline class is responsible
for managing and executing the various components (:py:class:`.PipelineComponent`)which make up
the processing of messages from and to users.
It provides a way to organize and structure the messages processing flow.
The Pipeline class is designed to be highly customizable and configurable,
allowing developers to add, remove, or modify the components that make up the messages processing flow.

The Pipeline class is designed to be used in conjunction with the :py:class:`.PipelineComponent`
class, which is defined in the Component module. Together, these classes provide a powerful and flexible way
to structure and manage the messages processing flow.
"""

import asyncio
import logging
from typing import Union, List, Dict, Optional, Hashable, Callable

from chatsky.context_storages import DBContextStorage
from chatsky.script import Script, Context, ActorStage
from chatsky.script import NodeLabel2Type, Message
from chatsky.utils.turn_caching import cache_clear

from chatsky.messengers.console import CLIMessengerInterface
from chatsky.messengers.common import MessengerInterface
from chatsky.slots.slots import GroupSlot
from ..service.group import ServiceGroup
from ..types import (
    ServiceBuilder,
    ServiceGroupBuilder,
    PipelineBuilder,
    GlobalExtraHandlerType,
    ExtraHandlerFunction,
    ExtraHandlerBuilder,
)
from .utils import finalize_service_group, pretty_format_component_info_dict
from chatsky.pipeline.pipeline.actor import Actor
from chatsky.pipeline.pipeline.script_parsing import JSONImporter, Path

logger = logging.getLogger(__name__)

ACTOR = "ACTOR"


class Pipeline:
    """
    Class that automates service execution and creates service pipeline.
    It accepts constructor parameters:

    :param components: (required) A :py:data:`~.ServiceGroupBuilder` object,
        that will be transformed to root service group. It should include :py:class:`~.Actor`,
        but only once (raises exception otherwise). It will always be named pipeline.
    :param script: (required) A :py:class:`~.Script` instance (object or dict).
    :param start_label: (required) Actor start label.
    :param fallback_label: Actor fallback label.
    :param label_priority: Default priority value for all actor :py:const:`labels <chatsky.script.ConstLabel>`
        where there is no priority. Defaults to `1.0`.
    :param condition_handler: Handler that processes a call of actor condition functions. Defaults to `None`.
    :param slots: Slots configuration.
    :param handlers: This variable is responsible for the usage of external handlers on
        the certain stages of work of :py:class:`~chatsky.script.Actor`.

        - key: :py:class:`~chatsky.script.ActorStage` - Stage in which the handler is called.
        - value: List[Callable] - The list of called handlers for each stage. Defaults to an empty `dict`.

    :param messenger_interface: An `AbsMessagingInterface` instance for this pipeline.
    :param context_storage: An :py:class:`~.DBContextStorage` instance for this pipeline or
        a dict to store dialog :py:class:`~.Context`.
    :param before_handler: List of `ExtraHandlerBuilder` to add to the group.
    :type before_handler: Optional[:py:data:`~.ExtraHandlerBuilder`]
    :param after_handler: List of `ExtraHandlerBuilder` to add to the group.
    :type after_handler: Optional[:py:data:`~.ExtraHandlerBuilder`]
    :param timeout: Timeout to add to pipeline root service group.
    :param optimization_warnings: Asynchronous pipeline optimization check request flag;
        warnings will be sent to logs. Additionally it has some calculated fields:

        - `_services_pipeline` is a pipeline root :py:class:`~.ServiceGroup` object,
        - `actor` is a pipeline actor, found among services.
    :param parallelize_processing: This flag determines whether or not the functions
        defined in the ``PRE_RESPONSE_PROCESSING`` and ``PRE_TRANSITIONS_PROCESSING`` sections
        of the script should be parallelized over respective groups.

    """

    def __init__(
        self,
        components: ServiceGroupBuilder,
        script: Union[Script, Dict],
        start_label: NodeLabel2Type,
        fallback_label: Optional[NodeLabel2Type] = None,
        label_priority: float = 1.0,
        condition_handler: Optional[Callable] = None,
        slots: Optional[Union[GroupSlot, Dict]] = None,
        handlers: Optional[Dict[ActorStage, List[Callable]]] = None,
        messenger_interface: Optional[MessengerInterface] = None,
        context_storage: Optional[Union[DBContextStorage, Dict]] = None,
        before_handler: Optional[ExtraHandlerBuilder] = None,
        after_handler: Optional[ExtraHandlerBuilder] = None,
        timeout: Optional[float] = None,
        optimization_warnings: bool = False,
        parallelize_processing: bool = False,
    ):
        self.actor: Actor = None
        self.messenger_interface = CLIMessengerInterface() if messenger_interface is None else messenger_interface
        self.context_storage = {} if context_storage is None else context_storage
        self.slots = GroupSlot.model_validate(slots) if slots is not None else None
        self._services_pipeline = ServiceGroup(
            components,
            before_handler=before_handler,
            after_handler=after_handler,
            timeout=timeout,
        )

        self._services_pipeline.name = "pipeline"
        self._services_pipeline.path = ".pipeline"
        actor_exists = finalize_service_group(self._services_pipeline, path=self._services_pipeline.path)
        if not actor_exists:
            raise Exception("Actor not found in the pipeline!")
        else:
            self.set_actor(
                script,
                start_label,
                fallback_label,
                label_priority,
                condition_handler,
                handlers,
            )
        if self.actor is None:
            raise Exception("Actor wasn't initialized correctly!")

        if optimization_warnings:
            self._services_pipeline.log_optimization_warnings()

        self.parallelize_processing = parallelize_processing

        # NB! The following API is highly experimental and may be removed at ANY time WITHOUT FURTHER NOTICE!!
        self._clean_turn_cache = True
        if self._clean_turn_cache:
            self.actor._clean_turn_cache = False

    def add_global_handler(
        self,
        global_handler_type: GlobalExtraHandlerType,
        extra_handler: ExtraHandlerFunction,
        whitelist: Optional[List[str]] = None,
        blacklist: Optional[List[str]] = None,
    ):
        """
        Method for adding global wrappers to pipeline.
        Different types of global wrappers are called before/after pipeline execution
        or before/after each pipeline component.
        They can be used for pipeline statistics collection or other functionality extensions.
        NB! Global wrappers are still wrappers,
        they shouldn't be used for much time-consuming tasks (see ../service/wrapper.py).

        :param global_handler_type: (required) indication where the wrapper
            function should be executed.
        :param extra_handler: (required) wrapper function itself.
        :type extra_handler: ExtraHandlerFunction
        :param whitelist: a list of services to only add this wrapper to.
        :param blacklist: a list of services to not add this wrapper to.
        :return: `None`
        """

        def condition(name: str) -> bool:
            return (whitelist is None or name in whitelist) and (blacklist is None or name not in blacklist)

        if (
            global_handler_type is GlobalExtraHandlerType.BEFORE_ALL
            or global_handler_type is GlobalExtraHandlerType.AFTER_ALL
        ):
            whitelist = ["pipeline"]
            global_handler_type = (
                GlobalExtraHandlerType.BEFORE
                if global_handler_type is GlobalExtraHandlerType.BEFORE_ALL
                else GlobalExtraHandlerType.AFTER
            )

        self._services_pipeline.add_extra_handler(global_handler_type, extra_handler, condition)

    @property
    def info_dict(self) -> dict:
        """
        Property for retrieving info dictionary about this pipeline.
        Returns info dict, containing most important component public fields as well as its type.
        All complex or unserializable fields here are replaced with 'Instance of [type]'.
        """
        return {
            "type": type(self).__name__,
            "messenger_interface": f"Instance of {type(self.messenger_interface).__name__}",
            "context_storage": f"Instance of {type(self.context_storage).__name__}",
            "services": [self._services_pipeline.info_dict],
        }

    def pretty_format(self, show_extra_handlers: bool = False, indent: int = 4) -> str:
        """
        Method for receiving pretty-formatted string description of the pipeline.
        Resulting string structure is somewhat similar to YAML string.
        Should be used in debugging/logging purposes and should not be parsed.

        :param show_wrappers: Whether to include Wrappers or not (could be many and/or generated).
        :param indent: Offset from new line to add before component children.
        """
        return pretty_format_component_info_dict(self.info_dict, show_extra_handlers, indent=indent)

    @classmethod
    def from_script(
        cls,
        script: Union[Script, Dict],
        start_label: NodeLabel2Type,
        fallback_label: Optional[NodeLabel2Type] = None,
        label_priority: float = 1.0,
        condition_handler: Optional[Callable] = None,
        slots: Optional[Union[GroupSlot, Dict]] = None,
        parallelize_processing: bool = False,
        handlers: Optional[Dict[ActorStage, List[Callable]]] = None,
        context_storage: Optional[Union[DBContextStorage, Dict]] = None,
        messenger_interface: Optional[MessengerInterface] = None,
        pre_services: Optional[List[Union[ServiceBuilder, ServiceGroupBuilder]]] = None,
        post_services: Optional[List[Union[ServiceBuilder, ServiceGroupBuilder]]] = None,
    ) -> "Pipeline":
        """
        Pipeline script-based constructor.
        It creates :py:class:`~.Actor` object and wraps it with pipeline.
        NB! It is generally not designed for projects with complex structure.
        :py:class:`~.Service` and :py:class:`~.ServiceGroup` customization
        becomes not as obvious as it could be with it.
        Should be preferred for simple workflows with Actor auto-execution.

        :param script: (required) A :py:class:`~.Script` instance (object or dict).
        :param start_label: (required) Actor start label.
        :param fallback_label: Actor fallback label.
        :param label_priority: Default priority value for all actor :py:const:`labels <chatsky.script.ConstLabel>`
            where there is no priority. Defaults to `1.0`.
        :param condition_handler: Handler that processes a call of actor condition functions. Defaults to `None`.
        :param slots: Slots configuration.
        :param parallelize_processing: This flag determines whether or not the functions
            defined in the ``PRE_RESPONSE_PROCESSING`` and ``PRE_TRANSITIONS_PROCESSING`` sections
            of the script should be parallelized over respective groups.
        :param handlers: This variable is responsible for the usage of external handlers on
            the certain stages of work of :py:class:`~chatsky.script.Actor`.

            - key: :py:class:`~chatsky.script.ActorStage` - Stage in which the handler is called.
            - value: List[Callable] - The list of called handlers for each stage. Defaults to an empty `dict`.

        :param context_storage: An :py:class:`~.DBContextStorage` instance for this pipeline
            or a dict to store dialog :py:class:`~.Context`.
        :param messenger_interface: An instance for this pipeline.
        :param pre_services: List of :py:data:`~.ServiceBuilder` or
            :py:data:`~.ServiceGroupBuilder` that will be executed before Actor.
        :type pre_services: Optional[List[Union[ServiceBuilder, ServiceGroupBuilder]]]
        :param post_services: List of :py:data:`~.ServiceBuilder` or
            :py:data:`~.ServiceGroupBuilder` that will be executed after Actor.
            It constructs root service group by merging `pre_services` + actor + `post_services`.
        :type post_services: Optional[List[Union[ServiceBuilder, ServiceGroupBuilder]]]
        """
        pre_services = [] if pre_services is None else pre_services
        post_services = [] if post_services is None else post_services
        return cls(
            script=script,
            start_label=start_label,
            fallback_label=fallback_label,
            label_priority=label_priority,
            condition_handler=condition_handler,
            slots=slots,
            parallelize_processing=parallelize_processing,
            handlers=handlers,
            messenger_interface=messenger_interface,
            context_storage=context_storage,
            components=[*pre_services, ACTOR, *post_services],
        )

    @classmethod
    def from_file(
        cls,
        file: Union[str, Path],
        # validation_stage: Optional[bool] = None,
        condition_handler: Optional[Callable] = None,
        # verbose: bool = True,
        parallelize_processing: bool = False,
        handlers: Optional[Dict[ActorStage, List[Callable]]] = None,
        context_storage: Optional[Union[DBContextStorage, Dict]] = None,
        messenger_interface: Optional[MessengerInterface] = None,
        pre_services: Optional[List[Union[ServiceBuilder, ServiceGroupBuilder]]] = None,
        post_services: Optional[List[Union[ServiceBuilder, ServiceGroupBuilder]]] = None,
    ):
        pre_services = [] if pre_services is None else pre_services
        post_services = [] if post_services is None else post_services
        script = JSONImporter(file).import_script()

        def to_tuple(i):
            if isinstance(i, list):
                return tuple(i)
            return i

        params = {
            param: to_tuple(script["CONFIG"].get(param))
            for param in script["CONFIG"].keys()
            if param != JSONImporter.CUSTOM_DIR_CONFIG_OPTION
        }
        del script["CONFIG"]  # todo: add support for CONFIG

        return cls(
            script=script,
            **params,
            condition_handler=condition_handler,
            parallelize_processing=parallelize_processing,
            handlers=handlers,
            messenger_interface=messenger_interface,
            context_storage=context_storage,
            components=[*pre_services, ACTOR, *post_services],
        )

    def set_actor(
        self,
        script: Union[Script, Dict],
        start_label: NodeLabel2Type,
        fallback_label: Optional[NodeLabel2Type] = None,
        label_priority: float = 1.0,
        condition_handler: Optional[Callable] = None,
        handlers: Optional[Dict[ActorStage, List[Callable]]] = None,
    ):
        """
        Set actor for the current pipeline and conducts necessary checks.
        Reset actor to previous if any errors are found.

        :param script: (required) A :py:class:`~.Script` instance (object or dict).
        :param start_label: (required) Actor start label.
            The start node of :py:class:`~chatsky.script.Script`. The execution begins with it.
        :param fallback_label: Actor fallback label. The label of :py:class:`~chatsky.script.Script`.
            Dialog comes into that label if all other transitions failed,
            or there was an error while executing the scenario.
        :param label_priority: Default priority value for all actor :py:const:`labels <chatsky.script.ConstLabel>`
            where there is no priority. Defaults to `1.0`.
        :param condition_handler: Handler that processes a call of actor condition functions. Defaults to `None`.
        :param handlers: This variable is responsible for the usage of external handlers on
            the certain stages of work of :py:class:`~chatsky.script.Actor`.

            - key :py:class:`~chatsky.script.ActorStage` - Stage in which the handler is called.
            - value List[Callable] - The list of called handlers for each stage. Defaults to an empty `dict`.
        """
        self.actor = Actor(script, start_label, fallback_label, label_priority, condition_handler, handlers)

    @classmethod
    def from_dict(cls, dictionary: PipelineBuilder) -> "Pipeline":
        """
        Pipeline dictionary-based constructor.
        Dictionary should have the fields defined in Pipeline main constructor,
        it will be split and passed to it as `**kwargs`.
        """
        return cls(**dictionary)

    async def _run_pipeline(
        self, request: Message, ctx_id: Optional[Hashable] = None, update_ctx_misc: Optional[dict] = None
    ) -> Context:
        """
        Method that should be invoked on user input.
        This method has the same signature as :py:class:`~chatsky.pipeline.types.PipelineRunnerFunction`.
        """
        if ctx_id is None:
            ctx = Context()
        elif isinstance(self.context_storage, DBContextStorage):
            ctx = await self.context_storage.get_async(ctx_id, Context(id=ctx_id))
        else:
            ctx = self.context_storage.get(ctx_id, Context(id=ctx_id))

        if update_ctx_misc is not None:
            ctx.misc.update(update_ctx_misc)

        if self.slots is not None:
            ctx.framework_data.slot_manager.set_root_slot(self.slots)

        ctx.add_request(request)
        result = await self._services_pipeline(ctx, self)

        if asyncio.iscoroutine(result):
            await result

        ctx.framework_data.service_states.clear()

        if isinstance(self.context_storage, DBContextStorage):
            await self.context_storage.set_item_async(ctx_id, ctx)
        else:
            self.context_storage[ctx_id] = ctx
        if self._clean_turn_cache:
            cache_clear()

        return ctx

    def run(self):
        """
        Method that starts a pipeline and connects to `messenger_interface`.
        It passes `_run_pipeline` to `messenger_interface` as a callbacks,
        so every time user request is received, `_run_pipeline` will be called.
        This method can be both blocking and non-blocking. It depends on current `messenger_interface` nature.
        Message interfaces that run in a loop block current thread.
        """
        asyncio.run(self.messenger_interface.connect(self._run_pipeline))

    def __call__(
        self, request: Message, ctx_id: Optional[Hashable] = None, update_ctx_misc: Optional[dict] = None
    ) -> Context:
        """
        Method that executes pipeline once.
        Basically, it is a shortcut for `_run_pipeline`.
        NB! When pipeline is executed this way, `messenger_interface` won't be initiated nor connected.

        This method has the same signature as :py:class:`~chatsky.pipeline.types.PipelineRunnerFunction`.
        """
        return asyncio.run(self._run_pipeline(request, ctx_id, update_ctx_misc))

    @property
    def script(self) -> Script:
        return self.actor.script



# -*- coding: utf-8 -*-
from chatsky.utils.db_benchmark.benchmark import (
    time_context_read_write,
    DBFactory,
    BenchmarkConfig,
    BenchmarkCase,
    save_results_to_file,
    benchmark_all,
)
from chatsky.utils.db_benchmark.report import report
from chatsky.utils.db_benchmark.basic_config import BasicBenchmarkConfig, basic_configurations



"""
Response comparer
-----------------
This module defines function used to compare two response objects.
"""

from typing import Any, Optional

from chatsky.script import Context, Message


def default_comparer(candidate: Message, reference: Message, _: Context) -> Optional[Any]:
    """
    The default response comparer. Literally compares two response objects.

    :param candidate: The received (candidate) response.
    :param reference: The true (reference) response.
    :param _: Current Context (unused).
    :return: `None` if two responses are equal or candidate response otherwise.
    """
    return None if candidate == reference else candidate



"""
Basic Config
------------
This module contains basic benchmark configurations.

It defines a simple configurations class (:py:class:`~.BasicBenchmarkConfig`)
as well as a set of configurations that covers different dialogs a user might have and some edge-cases
(:py:data:`~.basic_configurations`).
"""

from typing import Tuple, Optional
import string
import random

from humanize import naturalsize
from pympler import asizeof

from chatsky.script import Message, Context
from chatsky.utils.db_benchmark.benchmark import BenchmarkConfig


def get_dict(dimensions: Tuple[int, ...]):
    """
    Return misc dictionary build in `dimensions` dimensions.

    :param dimensions:
        Dimensions of the dictionary.
        Each element of the dimensions tuple is the number of keys on the corresponding level of the dictionary.
        The last element of the dimensions tuple is the length of the string values of the dict.

        e.g. dimensions=(1, 2) returns a dictionary with 1 key that points to a string of len 2.
        whereas dimensions=(1, 2, 3) returns a dictionary with 1 key that points to a dictionary
        with 2 keys each of which points to a string of len 3.

        So, the len of dimensions is the depth of the dictionary, while its values are
        the width of the dictionary at each level.
    """

    def _get_dict(dimensions: Tuple[int, ...]):
        if len(dimensions) < 2:
            # get a random string of length dimensions[0]
            return "".join(random.choice(string.printable) for _ in range(dimensions[0]))
        return {str(i): _get_dict(dimensions[1:]) for i in range(dimensions[0])}

    if len(dimensions) > 1:
        return _get_dict(dimensions)
    elif len(dimensions) == 1:
        return _get_dict((dimensions[0], 0))
    else:
        return _get_dict((0, 0))


def get_message(message_dimensions: Tuple[int, ...]):
    """
    Return message with a non-empty misc field.

    :param message_dimensions: Dimensions of the misc field of the message. See :py:func:`~.get_dict`.
    """
    return Message(misc=get_dict(message_dimensions))


def get_context(
    dialog_len: int,
    message_dimensions: Tuple[int, ...],
    misc_dimensions: Tuple[int, ...],
) -> Context:
    """
    Return context with a non-empty misc, labels, requests, responses fields.

    :param dialog_len: Number of labels, requests and responses.
    :param message_dimensions:
        A parameter used to generate messages for requests and responses. See :py:func:`~.get_message`.
    :param misc_dimensions:
        A parameter used to generate misc field. See :py:func:`~.get_dict`.
    """
    return Context(
        labels={i: (f"flow_{i}", f"node_{i}") for i in range(dialog_len)},
        requests={i: get_message(message_dimensions) for i in range(dialog_len)},
        responses={i: get_message(message_dimensions) for i in range(dialog_len)},
        misc=get_dict(misc_dimensions),
    )


class BasicBenchmarkConfig(BenchmarkConfig, frozen=True):
    """
    A simple benchmark configuration that generates contexts using two parameters:

    - `message_dimensions` -- to configure the way messages are generated.
    - `misc_dimensions` -- to configure size of context's misc field.

    Dialog length is configured using `from_dialog_len`, `to_dialog_len`, `step_dialog_len`.
    """

    context_num: int = 30
    """
    Number of times the contexts will be benchmarked.
    Increasing this number decreases standard error of the mean for benchmarked data.
    """
    from_dialog_len: int = 300
    """Starting dialog len of a context."""
    to_dialog_len: int = 311
    """
    Final dialog len of a context.
    :py:meth:`~.BasicBenchmarkConfig.context_updater` will return contexts
    until their dialog len is less then `to_dialog_len`.
    """
    step_dialog_len: int = 1
    """
    Increment step for dialog len.
    :py:meth:`~.BasicBenchmarkConfig.context_updater` will return contexts
    increasing dialog len by `step_dialog_len`.
    """
    message_dimensions: Tuple[int, ...] = (10, 10)
    """
    Dimensions of misc dictionaries inside messages.
    See :py:func:`~.get_message`.
    """
    misc_dimensions: Tuple[int, ...] = (10, 10)
    """
    Dimensions of misc dictionary.
    See :py:func:`~.get_dict`.
    """

    def get_context(self) -> Context:
        """
        Return context with `from_dialog_len`, `message_dimensions`, `misc_dimensions`.

        Wraps :py:func:`~.get_context`.
        """
        return get_context(self.from_dialog_len, self.message_dimensions, self.misc_dimensions)

    def info(self):
        """
        Return fields of this instance and sizes of objects defined by this config.

        :return:
            A dictionary with two keys.
            Key "params" stores fields of this configuration.
            Key "sizes" stores string representation of following values:

                - "starting_context_size" -- size of a context with `from_dialog_len`.
                - "final_context_size" -- size of a context with `to_dialog_len`.
                  A context of this size will never actually be benchmarked.
                - "misc_size" -- size of a misc field of a context.
                - "message_size" -- size of a misc field of a message.
        """
        return {
            "params": self.model_dump(),
            "sizes": {
                "starting_context_size": naturalsize(asizeof.asizeof(self.get_context()), gnu=True),
                "final_context_size": naturalsize(
                    asizeof.asizeof(get_context(self.to_dialog_len, self.message_dimensions, self.misc_dimensions)),
                    gnu=True,
                ),
                "misc_size": naturalsize(asizeof.asizeof(get_dict(self.misc_dimensions)), gnu=True),
                "message_size": naturalsize(asizeof.asizeof(get_message(self.message_dimensions)), gnu=True),
            },
        }

    def context_updater(self, context: Context) -> Optional[Context]:
        """
        Update context to have `step_dialog_len` more labels, requests and responses,
        unless such dialog len would be equal to `to_dialog_len` or exceed than it,
        in which case None is returned.
        """
        start_len = len(context.labels)
        if start_len + self.step_dialog_len < self.to_dialog_len:
            for i in range(start_len, start_len + self.step_dialog_len):
                context.add_label((f"flow_{i}", f"node_{i}"))
                context.add_request(get_message(self.message_dimensions))
                context.add_response(get_message(self.message_dimensions))
            return context
        else:
            return None


basic_configurations = {
    "large-misc": BasicBenchmarkConfig(
        from_dialog_len=1,
        to_dialog_len=50,
        message_dimensions=(3, 5, 6, 5, 3),
        misc_dimensions=(2, 4, 3, 8, 100),
    ),
    "short-messages": BasicBenchmarkConfig(
        from_dialog_len=500,
        to_dialog_len=550,
        message_dimensions=(2, 30),
        misc_dimensions=(0, 0),
    ),
    "default": BasicBenchmarkConfig(),
    "large-misc--long-dialog": BasicBenchmarkConfig(
        from_dialog_len=500,
        to_dialog_len=550,
        message_dimensions=(3, 5, 6, 5, 3),
        misc_dimensions=(2, 4, 3, 8, 100),
    ),
    "very-long-dialog-len": BasicBenchmarkConfig(
        context_num=10,
        from_dialog_len=10000,
        to_dialog_len=10050,
    ),
    "very-long-message-len": BasicBenchmarkConfig(
        context_num=10,
        from_dialog_len=1,
        to_dialog_len=3,
        message_dimensions=(10000, 1),
    ),
    "very-long-misc-len": BasicBenchmarkConfig(
        context_num=10,
        from_dialog_len=1,
        to_dialog_len=3,
        misc_dimensions=(10000, 1),
    ),
}
"""
Configuration that covers many dialog cases (as well as some edge-cases).

:meta hide-value:
"""



"""
Toy script
----------
This module contains a simple script and a dialog which are used
in tutorials.
"""

from chatsky.script.conditions import exact_match
from chatsky.script import TRANSITIONS, RESPONSE, Message

TOY_SCRIPT = {
    "greeting_flow": {
        "start_node": {
            RESPONSE: Message(),
            TRANSITIONS: {"node1": exact_match("Hi")},
        },
        "node1": {
            RESPONSE: Message("Hi, how are you?"),
            TRANSITIONS: {"node2": exact_match("i'm fine, how are you?")},
        },
        "node2": {
            RESPONSE: Message("Good. What do you want to talk about?"),
            TRANSITIONS: {"node3": exact_match("Let's talk about music.")},
        },
        "node3": {
            RESPONSE: Message("Sorry, I can not talk about music now."),
            TRANSITIONS: {"node4": exact_match("Ok, goodbye.")},
        },
        "node4": {RESPONSE: Message("bye"), TRANSITIONS: {"node1": exact_match("Hi")}},
        "fallback_node": {
            RESPONSE: Message("Ooops"),
            TRANSITIONS: {"node1": exact_match("Hi")},
        },
    }
}
"""
An example of a simple script.

:meta hide-value:
"""

TOY_SCRIPT_ARGS = (TOY_SCRIPT, ("greeting_flow", "start_node"), ("greeting_flow", "fallback_node"))
"""
Arguments to pass to :py:meth:`~chatsky.pipeline.pipeline.pipeline.Pipeline.from_script` in order to
use :py:data:`~.TOY_SCRIPT`:

.. code-block::

    Pipeline.from_script(*TOY_SCRIPT_ARGS, context_storage=..., ...)

:meta hide-value:
"""

HAPPY_PATH = (
    ("Hi", "Hi, how are you?"),
    ("i'm fine, how are you?", "Good. What do you want to talk about?"),
    ("Let's talk about music.", "Sorry, I can not talk about music now."),
    ("Ok, goodbye.", "bye"),
    ("Hi", "Hi, how are you?"),
)
"""
An example of a simple dialog.

:meta hide-value:
"""

MULTIFLOW_SCRIPT = {
    "root": {
        "start": {
            RESPONSE: Message("Hi"),
            TRANSITIONS: {
                ("small_talk", "ask_some_questions"): exact_match("hi"),
                ("animals", "have_pets"): exact_match("i like animals"),
                ("animals", "like_animals"): exact_match("let's talk about animals"),
                ("news", "what_news"): exact_match("let's talk about news"),
            },
        },
        "fallback": {RESPONSE: Message("Oops")},
    },
    "animals": {
        "have_pets": {
            RESPONSE: Message("do you have pets?"),
            TRANSITIONS: {"what_animal": exact_match("yes")},
        },
        "like_animals": {
            RESPONSE: Message("do you like it?"),
            TRANSITIONS: {"what_animal": exact_match("yes")},
        },
        "what_animal": {
            RESPONSE: Message("what animals do you have?"),
            TRANSITIONS: {
                "ask_about_color": exact_match("bird"),
                "ask_about_breed": exact_match("dog"),
            },
        },
        "ask_about_color": {RESPONSE: Message("what color is it")},
        "ask_about_breed": {
            RESPONSE: Message("what is this breed?"),
            TRANSITIONS: {
                "ask_about_breed": exact_match("pereat"),
                "tell_fact_about_breed": exact_match("bulldog"),
                "ask_about_training": exact_match("I don't know"),
            },
        },
        "tell_fact_about_breed": {
            RESPONSE: Message("Bulldogs appeared in England as specialized bull-baiting dogs. "),
        },
        "ask_about_training": {RESPONSE: Message("Do you train your dog? ")},
    },
    "news": {
        "what_news": {
            RESPONSE: Message("what kind of news do you prefer?"),
            TRANSITIONS: {
                "ask_about_science": exact_match("science"),
                "ask_about_sport": exact_match("sport"),
            },
        },
        "ask_about_science": {
            RESPONSE: Message("i got news about science, do you want to hear?"),
            TRANSITIONS: {
                "science_news": exact_match("yes"),
                ("small_talk", "ask_some_questions"): exact_match("let's change the topic"),
            },
        },
        "science_news": {
            RESPONSE: Message("This is science news"),
            TRANSITIONS: {
                "what_news": exact_match("ok"),
                ("small_talk", "ask_some_questions"): exact_match("let's change the topic"),
            },
        },
        "ask_about_sport": {
            RESPONSE: Message("i got news about sport, do you want to hear?"),
            TRANSITIONS: {
                "sport_news": exact_match("yes"),
                ("small_talk", "ask_some_questions"): exact_match("let's change the topic"),
            },
        },
        "sport_news": {
            RESPONSE: Message("This is sport news"),
            TRANSITIONS: {
                "what_news": exact_match("ok"),
                ("small_talk", "ask_some_questions"): exact_match("let's change the topic"),
            },
        },
    },
    "small_talk": {
        "ask_some_questions": {
            RESPONSE: Message("how are you"),
            TRANSITIONS: {
                "ask_talk_about": exact_match("fine"),
                ("animals", "like_animals"): exact_match("let's talk about animals"),
                ("news", "what_news"): exact_match("let's talk about news"),
            },
        },
        "ask_talk_about": {
            RESPONSE: Message("what do you want to talk about"),
            TRANSITIONS: {
                ("animals", "like_animals"): exact_match("dog"),
                ("news", "what_news"): exact_match("let's talk about news"),
            },
        },
    },
}
"""
Simple dialog with multiple flows.

:meta hide-value:
"""

MULTIFLOW_REQUEST_OPTIONS = {
    "root": {
        "start": [
            "hi",
            "i like animals",
            "let's talk about animals",
        ]
    },
    "animals": {
        "have_pets": ["yes"],
        "like_animals": ["yes"],
        "what_animal": ["bird", "dog"],
        "ask_about_breed": ["pereat", "bulldog", "I don't know"],
    },
    "news": {
        "what_news": ["science", "sport"],
        "ask_about_science": ["yes", "let's change the topic"],
        "science_news": ["ok", "let's change the topic"],
        "ask_about_sport": ["yes", "let's change the topic"],
        "sport_news": ["ok", "let's change the topic"],
    },
    "small_talk": {
        "ask_some_questions": [
            "fine",
            "let's talk about animals",
            "let's talk about news",
        ],
        "ask_talk_about": ["dog", "let's talk about news"],
    },
}
"""
Request options for automated client requests for :py:data:`~.MULTIFLOW_SCRIPT`.

:meta hide-value:
"""



# -*- coding: utf-8 -*-

from .interface import (
    MessengerInterface,
    MessengerInterfaceWithAttachments,
    PollingMessengerInterface,
    CallbackMessengerInterface,
)
from .types import PollingInterfaceLoopFunction



"""
Normalization
-------------
Normalization module is used to normalize all python objects and functions to a format
that is suitable for script and actor execution process.
This module contains a basic set of functions for normalizing data in a dialog script.
"""

from __future__ import annotations
import logging
from typing import Union, Callable, Optional, TYPE_CHECKING

from .keywords import Keywords
from .context import Context
from .types import ConstLabel, ConditionType, Label, LabelType
from .message import Message

if TYPE_CHECKING:
    from chatsky.pipeline.pipeline.pipeline import Pipeline

logger = logging.getLogger(__name__)


def normalize_label(label: Label, default_flow_label: LabelType = "") -> Label:
    """
    The function that is used for normalization of
    :py:const:`label <chatsky.script.Label>`.

    :param label: If label is Callable the function is wrapped into try/except
        and normalization is used on the result of the function call with the name label.
    :param default_flow_label: flow_label is used if label does not contain flow_label.
    :return: Result of the label normalization
    """
    if callable(label):

        def get_label_handler(ctx: Context, pipeline: Pipeline) -> Optional[ConstLabel]:
            try:
                new_label = label(ctx, pipeline)
                if new_label is None:
                    return None
                new_label = normalize_label(new_label, default_flow_label)
                flow_label, node_label, _ = new_label
                node = pipeline.script.get(flow_label, {}).get(node_label)
                if not node:
                    raise Exception(f"Unknown transitions {new_label} for pipeline.script={pipeline.script}")
                if node_label in [Keywords.LOCAL, Keywords.GLOBAL]:
                    raise Exception(f"Invalid transition: can't transition to {flow_label}:{node_label}")
            except Exception as exc:
                new_label = None
                logger.error(f"Exception {exc} of function {label}", exc_info=exc)
            return new_label

        return get_label_handler  # create wrap to get uniq key for dictionary
    elif isinstance(label, str) or isinstance(label, Keywords):
        return (default_flow_label, label, float("-inf"))
    elif isinstance(label, tuple) and len(label) == 2 and isinstance(label[-1], float):
        return (default_flow_label, label[0], label[-1])
    elif isinstance(label, tuple) and len(label) == 2 and isinstance(label[-1], str):
        flow_label = label[0] or default_flow_label
        return (flow_label, label[-1], float("-inf"))
    elif isinstance(label, tuple) and len(label) == 3:
        flow_label = label[0] or default_flow_label
        return (flow_label, label[1], label[2])
    else:
        raise TypeError(f"Label '{label!r}' is of incorrect type. It has to follow the `Label`:\n" f"{Label!r}")


def normalize_condition(condition: ConditionType) -> Callable[[Context, Pipeline], bool]:
    """
    The function that is used to normalize `condition`

    :param condition: Condition to normalize.
    :return: The function condition wrapped into the try/except.
    """
    if callable(condition):

        def callable_condition_handler(ctx: Context, pipeline: Pipeline) -> bool:
            try:
                return condition(ctx, pipeline)
            except Exception as exc:
                logger.error(f"Exception {exc} of function {condition}", exc_info=exc)
                return False

        return callable_condition_handler


def normalize_response(
    response: Optional[Union[Message, Callable[[Context, "Pipeline"], Message]]]
) -> Callable[[Context, "Pipeline"], Message]:
    """
    This function is used to normalize response. If the response is a Callable, it is returned, otherwise
    the response is wrapped in an asynchronous function and this function is returned.

    :param response: Response to normalize.
    :return: Function that returns callable response.
    """
    if callable(response):
        return response
    else:
        if response is None:
            result = Message()
        elif isinstance(response, Message):
            result = response
        else:
            raise TypeError(type(response))

        async def response_handler(ctx: Context, pipeline: Pipeline):
            return result

        return response_handler



from typing import Union, Optional, Sequence
import importlib
import importlib.util
import importlib.machinery
import sys
import logging
from pathlib import Path
import json
from inspect import ismodule

from pydantic import JsonValue
import yaml
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


logger = logging.getLogger(__name__)


class JSONImportError(Exception):
    __notes__ = ["Please read the guide on YAML-formatted scripts: url here"]  # todo: update placeholder string


class JSONImporter:
    CHATSKY_NAMESPACE_PREFIX = "chatsky"
    CUSTOM_DIR_NAMESPACE_PREFIX = "custom_dir"
    CUSTOM_DIR_CONFIG_OPTION = "custom_dir"
    TRANSITIONS_KEY = "TRANSITIONS"
    CONFIG_KEY = "CONFIG"
    TRANSITION_ITEM_KEYS = {"lbl", "cnd"}

    def __init__(self, file: Union[str, Path]):
        if isinstance(file, str):
            file = Path(file)

        if file.suffix == ".json":
            with open(file, "r") as fd:
                script = json.load(fd)
        elif file.suffix in (".yaml", ".yml"):
            with open(file, "r") as fd:
                script = yaml.load(fd, Loader=Loader)
        else:
            raise JSONImportError("File should have a `.json`, `.yaml` or `.yml` extension")
        logger.info(f"Loaded file {file}")

        self.script = script
        config = self.script.get(self.CONFIG_KEY)
        if not isinstance(config, dict):
            raise JSONImportError("Config is not found -- your script has to define a CONFIG dictionary")
        self.config = config

        custom_dir = config.get(self.CUSTOM_DIR_CONFIG_OPTION)
        if custom_dir is not None:
            if not isinstance(custom_dir, str):
                raise JSONImportError("custom_dir must be a string")
            custom_dir_path = Path(custom_dir)
            if not custom_dir_path.is_absolute():
                custom_dir_path = (file.parent / custom_dir_path).resolve(strict=False)

            if not custom_dir_path.exists():
                raise JSONImportError(f"Could not find directory {custom_dir_path}. custom_dir: {custom_dir}")

            logger.info(f"custom_dir set to {custom_dir_path}")

            self._custom_dir_stem = str(custom_dir_path.stem)
            self._custom_dir_location = str(custom_dir_path.parent)
        else:
            self._custom_dir_location = None
        self._custom_modules = {}

    @staticmethod
    def is_resolvable(value: str) -> bool:
        return value.startswith(JSONImporter.CHATSKY_NAMESPACE_PREFIX + ".") or\
               value.startswith(JSONImporter.CUSTOM_DIR_NAMESPACE_PREFIX + ".")

    def import_custom_module(self, module_name: str, paths: Optional[Sequence[str]] = None):
        if module_name in self._custom_modules:
            return self._custom_modules[module_name]

        if paths is None:
            if self._custom_dir_location is None:
                raise JSONImportError("custom_dir option must be set in order to use objects from it")
            paths = [self._custom_dir_location]

        parent_name, _, child_name = module_name.rpartition(".")

        if parent_name:
            parent_module = self.import_custom_module(parent_name, paths)

            paths = parent_module.__spec__.submodule_search_locations
        else:
            # root level import; replace `custom_dir` with actual module name
            if child_name != self.CUSTOM_DIR_NAMESPACE_PREFIX:
                raise RuntimeError(f"Trying to import from custom_dir while using wrong module_name: {child_name!r}")
            child_name = self._custom_dir_stem

        for finder in sys.meta_path:
            spec = finder.find_spec(child_name, paths)
            if spec is not None:
                break
        else:
            raise ModuleNotFoundError(f"No module named {child_name!r} at {paths!r}")

        module = importlib.util.module_from_spec(spec)
        self._custom_modules[module_name] = module
        spec.loader.exec_module(module)
        return module

    def resolve_target_object(self, obj: str):
        module_name, _, obj_name = obj.rpartition(".")

        if obj.startswith(self.CHATSKY_NAMESPACE_PREFIX):
            module = importlib.import_module(module_name)
        else:
            module = self.import_custom_module(module_name)
        return getattr(module, obj_name)

    def import_script(self):
        return self.replace_script_objects(self.script)

    def replace_obj(self, obj: JsonValue):
        if not isinstance(obj, dict):
            raise JSONImportError(f"Chatsky object has to be a dictionary: {obj}")
        keys = obj.keys()
        if len(keys) != 1:
            raise JSONImportError(f"Chatsky object has to have only 1 key: {obj.keys()}")
        key = keys.__iter__().__next__()
        logger.debug(f"Replacing object: {key}")
        target_obj = self.resolve_target_object(key)

        if target_obj is None:
            raise ImportError(f"Could not find object {key}")

        if not callable(target_obj):
            raise JSONImportError(f"Object `{key}` has to be callable")

        args = []
        kwargs = {}
        if isinstance(obj[key], dict):
            for k, v in obj[key].items():
                kwargs[k] = self.replace_script_objects(v)
        elif isinstance(obj[key], list):
            for item in obj[key]:
                args.append(self.replace_script_objects(item))
        elif obj[key] is not None:
            args.append(self.replace_script_objects(obj[key]))

        return target_obj(*args, **kwargs)

    def process_transitions(self, transition_list: list):
        if not isinstance(transition_list, list):
            raise JSONImportError(f"Transitions value should be a list of dictionaries, not {transition_list}")

        transitions = {}
        for item in transition_list:
            if not isinstance(item, dict):
                raise JSONImportError(f"Transition items have to be dictionaries, not {item}")
            if item.keys() != self.TRANSITION_ITEM_KEYS:
                raise JSONImportError(f"Transition items' keys have to be `lbl` and `cnd`, not {item.keys()}")

            lbl = self.replace_script_objects(item["lbl"])
            if isinstance(lbl, list):
                lbl = tuple(lbl)
            cnd = self.replace_script_objects(item["cnd"])

            if isinstance(lbl, tuple) and lbl in transitions:
                raise JSONImportError(f"Label {lbl} already exists in {transitions}")

            transitions[lbl] = cnd
        return transitions

    def replace_string_values(self, obj: JsonValue):
        if not isinstance(obj, str):
            raise JSONImportError(f"Obj {obj} has to be a string")
        if self.is_resolvable(obj):
            target_obj = self.resolve_target_object(obj)

            if target_obj is None:
                raise JSONImportError(f"Could not find object {obj}")

            return target_obj
        raise RuntimeError()

    def replace_script_objects(self, obj: JsonValue):
        if isinstance(obj, dict):
            keys = obj.keys()
            if len(keys) == 1:
                key = keys.__iter__().__next__()
                if self.is_resolvable(key):
                    return self.replace_obj(obj)

            return {k: (
                self.replace_script_objects(v) if k != self.TRANSITIONS_KEY else self.process_transitions(v)
            ) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self.replace_script_objects(item) for item in obj]
        elif isinstance(obj, str):
            if self.is_resolvable(obj):
                return self.replace_string_values(obj)
        return obj


def get_chatsky_objects():
    def get_objects_from_submodule(submodule_name: str, alias: Optional[str] = None):
        module = importlib.import_module(submodule_name)

        return {
            ".".join([alias or submodule_name, name]): obj
            for name, obj in module.__dict__.items() if not name.startswith("_") and not ismodule(obj)
        }

    return {
        **get_objects_from_submodule("chatsky.cnd"),
        **get_objects_from_submodule("chatsky.rsp"),
        **get_objects_from_submodule("chatsky.lbl"),
        **get_objects_from_submodule("chatsky.msg", "chatsky"),
    }



"""
Conditions
----------
The conditions module contains functions that can be used to determine whether the pipeline component to which they
are attached should be executed or not.
The standard set of them allows user to setup dependencies between pipeline components.
"""

from __future__ import annotations
from typing import Optional, TYPE_CHECKING

from chatsky.script import Context

from .types import (
    StartConditionCheckerFunction,
    ComponentExecutionState,
    StartConditionCheckerAggregationFunction,
)

if TYPE_CHECKING:
    from chatsky.pipeline.pipeline.pipeline import Pipeline


def always_start_condition(_: Context, __: Pipeline) -> bool:
    """
    Condition that always allows service execution. It's the default condition for all services.

    :param _: Current dialog context.
    :param __: Pipeline.
    """
    return True


def service_successful_condition(path: Optional[str] = None) -> StartConditionCheckerFunction:
    """
    Condition that allows service execution, only if the other service was executed successfully.
    Returns :py:data:`~.StartConditionCheckerFunction`.

    :param path: The path of the condition pipeline component.
    """

    def check_service_state(ctx: Context, _: Pipeline):
        state = ctx.framework_data.service_states.get(path, ComponentExecutionState.NOT_RUN)
        return ComponentExecutionState[state] == ComponentExecutionState.FINISHED

    return check_service_state


def not_condition(func: StartConditionCheckerFunction) -> StartConditionCheckerFunction:
    """
    Condition that returns opposite boolean value to the one returned by incoming function.
    Returns :py:data:`~.StartConditionCheckerFunction`.

    :param func: The function to return opposite of.
    """

    def not_function(ctx: Context, pipeline: Pipeline):
        return not func(ctx, pipeline)

    return not_function


def aggregate_condition(
    aggregator: StartConditionCheckerAggregationFunction, *functions: StartConditionCheckerFunction
) -> StartConditionCheckerFunction:
    """
    Condition that returns aggregated boolean value from all booleans returned by incoming functions.
    Returns :py:data:`~.StartConditionCheckerFunction`.

    :param aggregator: The function that accepts list of booleans and returns a single boolean.
    :param functions: Functions to aggregate.
    """

    def aggregation_function(ctx: Context, pipeline: Pipeline):
        return aggregator([func(ctx, pipeline) for func in functions])

    return aggregation_function


def all_condition(*functions: StartConditionCheckerFunction) -> StartConditionCheckerFunction:
    """
    Condition that returns `True` only if all incoming functions return `True`.
    Returns :py:data:`~.StartConditionCheckerFunction`.

    :param functions: Functions to aggregate.
    """
    return aggregate_condition(all, *functions)


def any_condition(*functions: StartConditionCheckerFunction) -> StartConditionCheckerFunction:
    """
    Condition that returns `True` if any of incoming functions returns `True`.
    Returns :py:data:`~.StartConditionCheckerFunction`.

    :param functions: Functions to aggregate.
    """
    return aggregate_condition(any, *functions)



"""
Default Extractors
------------------
This module includes a pool of default extractors
that you can use out of the box.

The default configuration for Superset dashboard leverages the data collected
by the extractors below. In order to use the default charts,
make sure that you include those functions in your pipeline.
Detailed examples can be found in the `tutorials` section.

"""

from datetime import datetime

from chatsky.script import Context
from chatsky.pipeline import ExtraHandlerRuntimeInfo, Pipeline
from .utils import get_extra_handler_name


async def get_current_label(ctx: Context, pipeline: Pipeline, info: ExtraHandlerRuntimeInfo):
    """
    Extract the current label on each turn.
    This function is required for running the dashboard with the default configuration.

    .. note::

        Preferrably, it needs to be invoked as `after_handler` of the `Actor` service.

    """
    last_label = ctx.last_label
    if last_label is None:
        last_label = pipeline.actor.start_label[:2]
    return {"flow": last_label[0], "node": last_label[1], "label": ": ".join(last_label)}


async def get_timing_before(ctx: Context, _, info: ExtraHandlerRuntimeInfo):
    """
    Store the pipeline component's start time inside the context.
    This function is required for running the dashboard with the default configuration.
    """
    start_time = datetime.now()
    ctx.framework_data.stats[get_extra_handler_name(info, "time")] = start_time


async def get_timing_after(ctx: Context, _, info: ExtraHandlerRuntimeInfo):  # noqa: F811
    """
    Extract the pipeline component's execution time.
    Requires :py:func:`~.get_timing_before` to be called previously in order to calculate the time.
    This function is required for running the dashboard with the default configuration.
    """
    start_time = ctx.framework_data.stats.pop(get_extra_handler_name(info, "time"), None)
    if start_time is None:
        return None
    data = {"execution_time": str(datetime.now() - start_time)}
    return data


async def get_last_response(ctx: Context, _, info: ExtraHandlerRuntimeInfo):
    """
    Extract the text of the last response in the current context.
    This handler is best used together with the `ACTOR` component.

    This function is required to enable charts that aggregate requests and responses.
    """
    data = {"last_response": ctx.last_response.text}
    return data


async def get_last_request(ctx: Context, _, info: ExtraHandlerRuntimeInfo):
    """
    Extract the text of the last request in the current context.
    This handler is best used together with the `ACTOR` component.

    This function is required to enable charts that aggregate requests and responses.
    """
    data = {"last_request": ctx.last_request.text}
    return data


__all__ = ["get_current_label", "get_timing_before", "get_timing_after", "get_last_request", "get_last_response"]
"""
List of exported functions.

:meta hide-avlue:
"""



"""
Extra field helpers
-------------------
Helpers for managing pydantic extra fields.
"""

from typing import List

from pydantic import BaseModel


def grab_extra_fields(attachment: BaseModel, extra_fields: List[str]):
    """
    Convenience method for passing attachment extras as named arguments to API functions.
    This might be useful for making sure no typos appear in code.
    Accepts a list of extra names and makes a dictionary of extras mathing these names.

    :param attachment: attachment whose extras will be used.
    :param extra_fields: list of extras that will be used.
    """

    return {extra_field: attachment.__pydantic_extra__.get(extra_field, None) for extra_field in extra_fields}



# -*- coding: utf-8 -*-

from .std_conditions import (
    exact_match,
    has_text,
    regexp,
    check_cond_seq,
    aggregate,
    any,
    all,
    negation,
    has_last_labels,
    true,
    false,
    agg,
    neg,
    has_callback_query,
)



"""
Instrumentor
-------------
This modules contains the :py:class:`~OtelInstrumentor` class that implements
Opentelemetry's `BaseInstrumentor` interface and allows for automated
instrumentation of Chatsky applications,
e.g. for automated logging and log export.

For detailed reference, see `~OtelInstrumentor` class.
"""

import asyncio
from typing import Collection, Optional

from wrapt import wrap_function_wrapper, decorator
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap
from opentelemetry.metrics import get_meter, get_meter_provider, Meter
from opentelemetry.trace import get_tracer, get_tracer_provider, Tracer
from opentelemetry._logs import get_logger, get_logger_provider, Logger, SeverityNumber
from opentelemetry.trace import SpanKind, Span
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk._logs import LoggerProvider, LogRecord
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter

from chatsky.script.core.context import get_last_index
from chatsky.stats.utils import (
    resource,
    get_extra_handler_name,
    set_logger_destination,
    set_meter_destination,
    set_tracer_destination,
)
from chatsky.stats import default_extractors


INSTRUMENTS = ["chatsky"]


class OtelInstrumentor(BaseInstrumentor):
    """
    Utility class for instrumenting Chatsky-related functions
    that implements the :py:class:`~BaseInstrumentor` interface.
    :py:meth:`~instrument` and :py:meth:`~uninstrument` methods
    are available to apply and revert the instrumentation effects,
    e.g. enable and disable logging at runtime.

    .. code-block::

        chatsky_instrumentor = OtelInstrumentor()
        chatsky_instrumentor.instrument()
        chatsky_instrumentor.uninstrument()

    Opentelemetry provider instances can be optionally passed to the class constructor.
    Otherwise, the global logger, tracer and meter providers are leveraged.

    The class implements the :py:meth:`~__call__` method, so that
    regular functions can be decorated using the class instance.

    .. code-block::

        @chatsky_instrumentor
        async def function(context, pipeline, runtime_info):
            ...

    :param logger_provider: Opentelemetry logger provider. Used to construct a logger instance.
    :param tracer_provider: Opentelemetry tracer provider. Used to construct a tracer instance.
    :parame meter_provider: Opentelemetry meter provider. Used to construct a meter instance.
    """

    def __init__(self, logger_provider=None, tracer_provider=None, meter_provider=None) -> None:
        super().__init__()
        self._logger_provider: Optional[LoggerProvider] = None
        self._tracer_provider: Optional[TracerProvider] = None
        self._meter_provider: Optional[MeterProvider] = None
        self._logger: Optional[Logger] = None
        self._tracer: Optional[Tracer] = None
        self._meter: Optional[Meter] = None
        self._configure_providers(
            logger_provider=logger_provider, tracer_provider=tracer_provider, meter_provider=meter_provider
        )

    def __enter__(self):
        if not self.is_instrumented_by_opentelemetry:
            self.instrument()
        return self

    def __exit__(self):
        if self.is_instrumented_by_opentelemetry:
            self.uninstrument()

    @classmethod
    def from_url(cls, url: str, insecure: bool = True, timeout: Optional[int] = None):
        """
        Construct an instrumentor instance using only the url of the OTLP Collector.
        Inherently modifies the global provider instances adding an export destination
        for the target url.

        .. code-block::

            instrumentor = OtelInstrumentor.from_url("grpc://localhost:4317")

        :param url: Url of the running Otel Collector server. Due to limited support of HTTP protocol
            by the Opentelemetry Python extension, GRPC protocol is preferred.
        :param insecure: Whether non-SSL-protected connection is allowed. Defaults to True.
        :param timeout: Connection timeout in seconds, optional.
        """
        set_logger_destination(OTLPLogExporter(endpoint=url, insecure=insecure, timeout=timeout))
        set_tracer_destination(OTLPSpanExporter(endpoint=url, insecure=insecure, timeout=timeout))
        set_meter_destination(OTLPMetricExporter(endpoint=url, insecure=insecure, timeout=timeout))
        return cls()

    def instrumentation_dependencies(self) -> Collection[str]:
        """
        :meta private:

        Required libraries. Implements the Python Opentelemetry instrumentor interface.

        """
        return INSTRUMENTS

    def _instrument(self, logger_provider=None, tracer_provider=None, meter_provider=None):
        if any([logger_provider, meter_provider, tracer_provider]):
            self._configure_providers(
                logger_provider=logger_provider, tracer_provider=tracer_provider, meter_provider=meter_provider
            )
        for func_name in default_extractors.__all__:
            wrap_function_wrapper(default_extractors, func_name, self.__call__.__wrapped__)

    def _uninstrument(self, **kwargs):
        for func_name in default_extractors.__all__:
            unwrap(default_extractors, func_name)

    def _configure_providers(self, logger_provider, tracer_provider, meter_provider):
        self._logger_provider = logger_provider or get_logger_provider()
        self._tracer_provider = tracer_provider or get_tracer_provider()
        self._meter_provider = meter_provider or get_meter_provider()
        self._logger = get_logger(__name__, None, self._logger_provider)
        self._tracer = get_tracer(__name__, None, self._tracer_provider)
        self._meter = get_meter(__name__, None, self._meter_provider)

    @decorator
    async def __call__(self, wrapped, _, args, kwargs):
        """
        Regular functions that match the :py:class:`~chatsky.pipeline.types.ExtraHandlerFunction`
        signature can be decorated with the class instance to log the returned value.
        This method implements the logging procedure.
        The returned value is assumed to be `dict` or `NoneType`.
        Logging non-atomic values is discouraged, as they cannot be translated using
        the `Protobuf` protocol.
        Logging is ignored if the application is in 'uninstrumented' state.

        :param wrapped: Function to decorate.
        :param args: Positional arguments of the decorated function.
        :param kwargs: Keyword arguments of the decorated function.
        """
        ctx, _, info = args
        pipeline_component = get_extra_handler_name(info)
        attributes = {
            "context_id": str(ctx.id),
            "request_id": get_last_index(ctx.requests),
            "pipeline_component": pipeline_component,
        }

        result: Optional[dict]
        if asyncio.iscoroutinefunction(wrapped):
            result = await wrapped(ctx, _, info)
        else:
            result = wrapped(ctx, _, info)

        if result is None or not self.is_instrumented_by_opentelemetry:
            # self.is_instrumented_by_opentelemetry allows to disable
            # the decorator programmatically if
            # instrumentation is disabled.
            return result

        span: Span
        with self._tracer.start_as_current_span(wrapped.__name__, kind=SpanKind.INTERNAL) as span:
            span_ctx = span.get_span_context()
            record = LogRecord(
                span_id=span_ctx.span_id,
                trace_id=span_ctx.trace_id,
                body=result,
                trace_flags=span_ctx.trace_flags,
                severity_text=None,
                severity_number=SeverityNumber(1),
                resource=resource,
                attributes=attributes,
            )
            self._logger.emit(record=record)
        return result



"""
Serialization
-------------
Tools that provide JSON serialization via Pickle for unserializable objects.

- :py:data:`~.PickleEncodedValue`:
    A field annotated with this will be pickled/unpickled during JSON-serialization/validation.
- :py:data:`~.JSONSerializableDict`:
    A dictionary field annotated with this will make all its items smart-serializable:
    If an item is serializable -- nothing would change.
    Otherwise -- it will be serialized via pickle.
- :py:class:`~.JSONSerializableExtras`:
    A pydantic base class that makes its extra fields a `JSONSerializableDict`.
"""

from base64 import decodebytes, encodebytes
from copy import deepcopy
from pickle import dumps, loads
from typing import Any, Dict, List, Union
from typing_extensions import Annotated, TypeAlias
from pydantic import (
    JsonValue,
    PlainSerializer,
    PlainValidator,
    RootModel,
    BaseModel,
    model_validator,
    model_serializer,
)
from pydantic_core import PydanticSerializationError

_JSON_EXTRA_FIELDS_KEYS = "__pickled_extra_fields__"
"""
This key is used in :py:data:`~.JSONSerializableDict` to remember pickled items.
"""

Serializable: TypeAlias = Dict[str, Union[JsonValue, List[Any], Dict[str, Any], Any]]
"""Type annotation for objects supported by :py:func:`~.json_pickle_serializer`."""


class _WrapperModel(RootModel):
    """
    Wrapper model for testing whether an object is serializable to JSON.
    """

    root: Any


def pickle_serializer(value: Any) -> str:
    """
    Serializer function that serializes any pickle-serializable value into JSON-serializable.
    Serializes value with pickle and encodes bytes as base64 string.

    :param value: Pickle-serializable object.
    :return: String-encoded object.
    """

    return encodebytes(dumps(value)).decode()


def pickle_validator(value: str) -> Any:
    """
    Validator function that validates base64 string encoded bytes as a pickle-serializable value.
    Decodes base64 string and validates value with pickle.

    :param value: String-encoded string.
    :return: Pickle-serializable object.
    """

    return loads(decodebytes(value.encode()))


def json_pickle_serializer(model: Serializable) -> Serializable:
    """
    Serializer function that serializes a dictionary or Pydantic object to JSON.
    For every object field, it checks whether the field is JSON serializable,
    and if it's not, serializes it using pickle.
    It also keeps track of pickle-serializable field names in a special list.

    :param model: Pydantic model object or a dictionary.
    :original_serializer: Original serializer function for model.
    :return: model with all the fields serialized to JSON.
    """

    extra_fields = list()
    model_copy = deepcopy(model)

    for field_name, field_value in model_copy.items():
        try:
            if isinstance(field_value, bytes):
                raise PydanticSerializationError("")
            else:
                model_copy[field_name] = _WrapperModel(root=field_value).model_dump(mode="json")
        except PydanticSerializationError:
            model_copy[field_name] = pickle_serializer(field_value)
            extra_fields += [field_name]

    if len(extra_fields) > 0:
        model_copy[_JSON_EXTRA_FIELDS_KEYS] = extra_fields
    return model_copy


def json_pickle_validator(model: Serializable) -> Serializable:
    """
    Validator function that validates a JSON dictionary to a python dictionary.
    For every object field, it checks if the field is pickle-serialized,
    and if it is, validates it using pickle.

    :param model: Pydantic model object or a dictionary.
    :return: model with all the fields serialized to JSON.
    """

    model_copy = deepcopy(model)

    if _JSON_EXTRA_FIELDS_KEYS in model.keys():
        for extra_key in model[_JSON_EXTRA_FIELDS_KEYS]:
            extra_value = model[extra_key]
            model_copy[extra_key] = pickle_validator(extra_value)
        del model_copy[_JSON_EXTRA_FIELDS_KEYS]

    return model_copy


PickleSerializer = PlainSerializer(pickle_serializer, when_used="json")
"""Pydantic wrapper of :py:func:`~.pickle_serializer`."""
PickleValidator = PlainValidator(pickle_validator)
"""Pydantic wrapper of :py:func:`~.pickle_validator`."""
PickleEncodedValue = Annotated[Any, PickleSerializer, PickleValidator]
"""
Annotation for field that makes it JSON serializable via pickle:

This field is always a normal object when inside its class but is a string encoding of the object
outside of the class -- either after serialization or before initialization.
As such this field cannot be used during initialization and the only way to use it is to bypass validation.

.. code:: python

    class MyClass(BaseModel):
        my_field: Optional[PickleEncodedValue] = None  # the field must have a default value

    my_obj = MyClass()  # the field cannot be set during init
    my_obj.my_field = unserializable_object  # can be set manually to avoid validation

Alternatively, ``BaseModel.model_construct`` may be used to bypass validation,
though it would bypass validation of all fields.
"""

JSONPickleSerializer = PlainSerializer(json_pickle_serializer, when_used="json")
"""Pydantic wrapper of :py:func:`~.json_pickle_serializer`."""
JSONPickleValidator = PlainValidator(json_pickle_validator)
"""Pydantic wrapper of :py:func:`~.json_pickle_validator`."""
JSONSerializableDict = Annotated[Serializable, JSONPickleSerializer, JSONPickleValidator]
"""
Annotation for dictionary or Pydantic model that makes all its fields JSON serializable.

This uses a reserved dictionary key :py:data:`~._JSON_EXTRA_FIELDS_KEYS` to store
fields serialized that way.
"""


class JSONSerializableExtras(BaseModel, extra="allow"):
    """
    This model makes extra fields pickle-serializable.
    Do not use :py:data:`~._JSON_EXTRA_FIELDS_KEYS` as an extra field name.
    """

    def __init__(self, **kwargs):  # supress unknown arg warnings
        super().__init__(**kwargs)

    @model_validator(mode="after")
    def extra_validator(self):
        """
        Validate model along with the `extras` field: i.e. all the fields not listed in the model.

        :return: Validated model.
        """
        self.__pydantic_extra__ = json_pickle_validator(self.__pydantic_extra__)
        return self

    @model_serializer(mode="wrap", when_used="json")
    def extra_serializer(self, original_serializer) -> Dict[str, Any]:
        """
        Serialize model along with the `extras` field: i.e. all the fields not listed in the model.

        :param original_serializer: Function originally used for serialization by Pydantic.
        :return: Serialized model.
        """
        model_copy = self.model_copy(deep=True)
        for extra_name in self.model_extra.keys():
            delattr(model_copy, extra_name)
        model_dict = original_serializer(model_copy)
        model_dict.update(json_pickle_serializer(self.model_extra))
        return model_dict



# -*- coding: utf-8 -*-

from .core.context import Context
from .core.keywords import (
    Keywords,
    GLOBAL,
    LOCAL,
    TRANSITIONS,
    RESPONSE,
    MISC,
    PRE_RESPONSE_PROCESSING,
    PRE_TRANSITIONS_PROCESSING,
)
from .core.script import Node, Script
from .core.types import (
    LabelType,
    NodeLabel1Type,
    NodeLabel2Type,
    NodeLabel3Type,
    NodeLabelTupledType,
    ConstLabel,
    Label,
    ConditionType,
    ActorStage,
)
from .core.message import Message



"""
Extra Handler
-------------
The Extra Handler module contains additional functionality that extends the capabilities of the system
beyond the core functionality. Extra handlers is an input converting addition to :py:class:`.PipelineComponent`.
For example, it is used to grep statistics from components, timing, logging, etc.
"""

from __future__ import annotations
import asyncio
import logging
import inspect
from typing import Optional, List, TYPE_CHECKING

from chatsky.script import Context

from .utils import collect_defined_constructor_parameters_to_dict, _get_attrs_with_updates
from chatsky.utils.devel.async_helpers import wrap_sync_function_in_async
from ..types import (
    ServiceRuntimeInfo,
    ExtraHandlerType,
    ExtraHandlerBuilder,
    ExtraHandlerFunction,
    ExtraHandlerRuntimeInfo,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from chatsky.pipeline.pipeline.pipeline import Pipeline


class _ComponentExtraHandler:
    """
    Class, representing an extra pipeline component handler.
    A component extra handler is a set of functions, attached to pipeline component (before or after it).
    Extra handlers should execute supportive tasks (like time or resources measurement, minor data transformations).
    Extra handlers should NOT edit context or pipeline, use services for that purpose instead.

    :param functions: An `ExtraHandlerBuilder` object, an `_ComponentExtraHandler` instance,
        a dict or a list of :py:data:`~.ExtraHandlerFunction`.
    :type functions: :py:data:`~.ExtraHandlerBuilder`
    :param stage: An :py:class:`~.ExtraHandlerType`, specifying whether this handler will be executed before or
        after pipeline component.
    :param timeout: (for asynchronous only!) Maximum component execution time (in seconds),
        if it exceeds this time, it is interrupted.
    :param asynchronous: Requested asynchronous property.
    """

    def __init__(
        self,
        functions: ExtraHandlerBuilder,
        stage: ExtraHandlerType = ExtraHandlerType.UNDEFINED,
        timeout: Optional[float] = None,
        asynchronous: Optional[bool] = None,
    ):
        overridden_parameters = collect_defined_constructor_parameters_to_dict(
            timeout=timeout, asynchronous=asynchronous
        )
        if isinstance(functions, _ComponentExtraHandler):
            self.__init__(
                **_get_attrs_with_updates(
                    functions,
                    ("calculated_async_flag", "stage"),
                    {"requested_async_flag": "asynchronous"},
                    overridden_parameters,
                )
            )
        elif isinstance(functions, dict):
            functions.update(overridden_parameters)
            self.__init__(**functions)
        elif isinstance(functions, List):
            self.functions = functions
            self.timeout = timeout
            self.requested_async_flag = asynchronous
            self.calculated_async_flag = all([asyncio.iscoroutinefunction(func) for func in self.functions])
            self.stage = stage
        else:
            raise Exception(f"Unknown type for {type(self).__name__} {functions}")

    @property
    def asynchronous(self) -> bool:
        """
        Property, that indicates, whether this component extra handler is synchronous or asynchronous.
        It is calculated according to the following rules:

        - | If component **can** be asynchronous and :py:attr:`~requested_async_flag` is set,
            it returns :py:attr:`~requested_async_flag`.
        - | If component **can** be asynchronous and :py:attr:`~requested_async_flag` isn't set,
            it returns `True`.
        - | If component **can't** be asynchronous and :py:attr:`~requested_async_flag` is `False` or not set,
            it returns `False`.
        - | If component **can't** be asynchronous and :py:attr:`~requested_async_flag` is `True`,
            an Exception is thrown in constructor.

        """
        return self.calculated_async_flag if self.requested_async_flag is None else self.requested_async_flag

    async def _run_function(
        self, func: ExtraHandlerFunction, ctx: Context, pipeline: Pipeline, component_info: ServiceRuntimeInfo
    ):
        handler_params = len(inspect.signature(func).parameters)
        if handler_params == 1:
            await wrap_sync_function_in_async(func, ctx)
        elif handler_params == 2:
            await wrap_sync_function_in_async(func, ctx, pipeline)
        elif handler_params == 3:
            extra_handler_runtime_info = ExtraHandlerRuntimeInfo(func=func, stage=self.stage, component=component_info)
            await wrap_sync_function_in_async(func, ctx, pipeline, extra_handler_runtime_info)
        else:
            raise Exception(
                f"Too many parameters required for component {component_info.name} {self.stage}"
                f" wrapper handler '{func.__name__}': {handler_params}!"
            )

    async def _run(self, ctx: Context, pipeline: Pipeline, component_info: ServiceRuntimeInfo):
        """
        Method for executing one of the extra handler functions (before or after).
        If the function is not set, nothing happens.

        :param ctx: current dialog context.
        :param pipeline: the current pipeline.
        :param component_info: associated component's info dictionary.
        :return: `None`
        """

        if self.asynchronous:
            futures = [self._run_function(func, ctx, pipeline, component_info) for func in self.functions]
            for func, future in zip(self.functions, asyncio.as_completed(futures)):
                try:
                    await future
                except asyncio.TimeoutError:
                    logger.warning(f"Component {component_info.name} {self.stage} wrapper '{func.__name__}' timed out!")

        else:
            for func in self.functions:
                await self._run_function(func, ctx, pipeline, component_info)

    async def __call__(self, ctx: Context, pipeline: Pipeline, component_info: ServiceRuntimeInfo):
        """
        A method for calling pipeline components.
        It sets up timeout if this component is asynchronous and executes it using `_run` method.

        :param ctx: (required) Current dialog `Context`.
        :param pipeline: This `Pipeline`.
        :return: `Context` if this is a synchronous service or
            `Awaitable` if this is an asynchronous component or `None`.
        """
        if self.asynchronous:
            task = asyncio.create_task(self._run(ctx, pipeline, component_info))
            return await asyncio.wait_for(task, timeout=self.timeout)
        else:
            return await self._run(ctx, pipeline, component_info)

    @property
    def info_dict(self) -> dict:
        """
        Property for retrieving info dictionary about this extra handler.

        :return: Info dict, containing its fields as well as its type.
            All not set fields there are replaced with `None`.
        """
        return {
            "type": type(self).__name__,
            "timeout": self.timeout,
            "asynchronous": self.asynchronous,
            "functions": [func.__name__ for func in self.functions],
        }


class BeforeHandler(_ComponentExtraHandler):
    """
    A handler for extra functions that are executed before the component's main function.

    :param functions: A callable or a list of callables that will be executed
        before the component's main function.
    :type functions: ExtraHandlerBuilder
    :param timeout: Optional timeout for the execution of the extra functions, in
        seconds.
    :param asynchronous: Optional flag that indicates whether the extra functions
        should be executed asynchronously. The default value of the flag is True
        if all the functions in this handler are asynchronous.
    """

    def __init__(
        self,
        functions: ExtraHandlerBuilder,
        timeout: Optional[int] = None,
        asynchronous: Optional[bool] = None,
    ):
        super().__init__(functions, ExtraHandlerType.BEFORE, timeout, asynchronous)


class AfterHandler(_ComponentExtraHandler):
    """
    A handler for extra functions that are executed after the component's main function.

    :param functions: A callable or a list of callables that will be executed
        after the component's main function.
    :type functions: ExtraHandlerBuilder
    :param timeout: Optional timeout for the execution of the extra functions, in
        seconds.
    :param asynchronous: Optional flag that indicates whether the extra functions
        should be executed asynchronously. The default value of the flag is True
        if all the functions in this handler are asynchronous.
    """

    def __init__(
        self,
        functions: ExtraHandlerBuilder,
        timeout: Optional[int] = None,
        asynchronous: Optional[bool] = None,
    ):
        super().__init__(functions, ExtraHandlerType.AFTER, timeout, asynchronous)



"""
Service
-------
The Service module contains the :py:class:`.Service` class,
which can be included into pipeline as object or a dictionary.
Pipeline consists of services and service groups.
Service group can be synchronous or asynchronous.
Service is an atomic part of a pipeline.
Service can be asynchronous only if its handler is a coroutine.
Actor wrapping service is asynchronous.
"""

from __future__ import annotations
import logging
import inspect
from typing import Optional, TYPE_CHECKING

from chatsky.script import Context

from .utils import collect_defined_constructor_parameters_to_dict, _get_attrs_with_updates
from chatsky.utils.devel.async_helpers import wrap_sync_function_in_async
from ..types import (
    ServiceBuilder,
    StartConditionCheckerFunction,
    ComponentExecutionState,
    ExtraHandlerBuilder,
    ExtraHandlerType,
)
from ..pipeline.component import PipelineComponent

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from chatsky.pipeline.pipeline.pipeline import Pipeline


class Service(PipelineComponent):
    """
    This class represents a service.
    Service can be included into pipeline as object or a dictionary.
    Service group can be synchronous or asynchronous.
    Service can be asynchronous only if its handler is a coroutine.

    :param handler: A service function or an actor.
    :type handler: :py:data:`~.ServiceBuilder`
    :param before_handler: List of `ExtraHandlerBuilder` to add to the group.
    :type before_handler: Optional[:py:data:`~.ExtraHandlerBuilder`]
    :param after_handler: List of `ExtraHandlerBuilder` to add to the group.
    :type after_handler: Optional[:py:data:`~.ExtraHandlerBuilder`]
    :param timeout: Timeout to add to the group.
    :param asynchronous: Requested asynchronous property.
    :param start_condition: StartConditionCheckerFunction that is invoked before each service execution;
        service is executed only if it returns `True`.
    :type start_condition: Optional[:py:data:`~.StartConditionCheckerFunction`]
    :param name: Requested service name.
    """

    def __init__(
        self,
        handler: ServiceBuilder,
        before_handler: Optional[ExtraHandlerBuilder] = None,
        after_handler: Optional[ExtraHandlerBuilder] = None,
        timeout: Optional[float] = None,
        asynchronous: Optional[bool] = None,
        start_condition: Optional[StartConditionCheckerFunction] = None,
        name: Optional[str] = None,
    ):
        overridden_parameters = collect_defined_constructor_parameters_to_dict(
            before_handler=before_handler,
            after_handler=after_handler,
            timeout=timeout,
            asynchronous=asynchronous,
            start_condition=start_condition,
            name=name,
        )
        if isinstance(handler, dict):
            handler.update(overridden_parameters)
            self.__init__(**handler)
        elif isinstance(handler, Service):
            self.__init__(
                **_get_attrs_with_updates(
                    handler,
                    (
                        "calculated_async_flag",
                        "path",
                    ),
                    {"requested_async_flag": "asynchronous"},
                    overridden_parameters,
                )
            )
        elif callable(handler) or isinstance(handler, str) and handler == "ACTOR":
            self.handler = handler
            super(Service, self).__init__(
                before_handler,
                after_handler,
                timeout,
                True,
                True,
                start_condition,
                name,
            )
        else:
            raise Exception(f"Unknown type of service handler: {handler}")

    async def _run_handler(self, ctx: Context, pipeline: Pipeline) -> None:
        """
        Method for service `handler` execution.
        Handler has three possible signatures, so this method picks the right one to invoke.
        These possible signatures are:

        - (ctx: Context) - accepts current dialog context only.
        - (ctx: Context, pipeline: Pipeline) - accepts context and current pipeline.
        - | (ctx: Context, pipeline: Pipeline, info: ServiceRuntimeInfo) - accepts context,
              pipeline and service runtime info dictionary.

        :param ctx: Current dialog context.
        :param pipeline: The current pipeline.
        :return: `None`
        """
        handler_params = len(inspect.signature(self.handler).parameters)
        if handler_params == 1:
            await wrap_sync_function_in_async(self.handler, ctx)
        elif handler_params == 2:
            await wrap_sync_function_in_async(self.handler, ctx, pipeline)
        elif handler_params == 3:
            await wrap_sync_function_in_async(self.handler, ctx, pipeline, self._get_runtime_info(ctx))
        else:
            raise Exception(f"Too many parameters required for service '{self.name}' handler: {handler_params}!")

    async def _run_as_actor(self, ctx: Context, pipeline: Pipeline) -> None:
        """
        Method for running this service if its handler is an `Actor`.
        Catches runtime exceptions.

        :param ctx: Current dialog context.
        """
        try:
            await pipeline.actor(pipeline, ctx)
            self._set_state(ctx, ComponentExecutionState.FINISHED)
        except Exception as exc:
            self._set_state(ctx, ComponentExecutionState.FAILED)
            logger.error(f"Actor '{self.name}' execution failed!", exc_info=exc)

    async def _run_as_service(self, ctx: Context, pipeline: Pipeline) -> None:
        """
        Method for running this service if its handler is not an Actor.
        Checks start condition and catches runtime exceptions.

        :param ctx: Current dialog context.
        :param pipeline: Current pipeline.
        """
        try:
            if self.start_condition(ctx, pipeline):
                self._set_state(ctx, ComponentExecutionState.RUNNING)
                await self._run_handler(ctx, pipeline)
                self._set_state(ctx, ComponentExecutionState.FINISHED)
            else:
                self._set_state(ctx, ComponentExecutionState.NOT_RUN)
        except Exception as exc:
            self._set_state(ctx, ComponentExecutionState.FAILED)
            logger.error(f"Service '{self.name}' execution failed!", exc_info=exc)

    async def _run(self, ctx: Context, pipeline: Pipeline) -> None:
        """
        Method for handling this service execution.
        Executes extra handlers before and after execution, launches `_run_as_actor` or `_run_as_service` method.

        :param ctx: (required) Current dialog context.
        :param pipeline: the current pipeline.
        """
        await self.run_extra_handler(ExtraHandlerType.BEFORE, ctx, pipeline)

        if isinstance(self.handler, str) and self.handler == "ACTOR":
            await self._run_as_actor(ctx, pipeline)
        else:
            await self._run_as_service(ctx, pipeline)

        await self.run_extra_handler(ExtraHandlerType.AFTER, ctx, pipeline)

    @property
    def info_dict(self) -> dict:
        """
        See `Component.info_dict` property.
        Adds `handler` key to base info dictionary.
        """
        representation = super(Service, self).info_dict
        if isinstance(self.handler, str) and self.handler == "ACTOR":
            service_representation = "Instance of Actor"
        elif callable(self.handler):
            service_representation = f"Callable '{self.handler.__name__}'"
        else:
            service_representation = "[Unknown]"
        representation.update({"handler": service_representation})
        return representation


def to_service(
    before_handler: Optional[ExtraHandlerBuilder] = None,
    after_handler: Optional[ExtraHandlerBuilder] = None,
    timeout: Optional[int] = None,
    asynchronous: Optional[bool] = None,
    start_condition: Optional[StartConditionCheckerFunction] = None,
    name: Optional[str] = None,
):
    """
    Function for decorating a function as a Service.
    Returns a Service, constructed from this function (taken as a handler).
    All arguments are passed directly to `Service` constructor.
    """

    def inner(handler: ServiceBuilder) -> Service:
        return Service(
            handler=handler,
            before_handler=before_handler,
            after_handler=after_handler,
            timeout=timeout,
            asynchronous=asynchronous,
            start_condition=start_condition,
            name=name,
        )

    return inner



# -*- coding: utf-8 -*-


from .conditions import (
    always_start_condition,
    service_successful_condition,
    not_condition,
    aggregate_condition,
    all_condition,
    any_condition,
)
from .types import (
    ComponentExecutionState,
    GlobalExtraHandlerType,
    ExtraHandlerType,
    StartConditionCheckerFunction,
    StartConditionCheckerAggregationFunction,
    ExtraHandlerConditionFunction,
    ServiceRuntimeInfo,
    ExtraHandlerRuntimeInfo,
    ExtraHandlerFunction,
    ServiceFunction,
    ExtraHandlerBuilder,
    ServiceBuilder,
    ServiceGroupBuilder,
    PipelineBuilder,
)

from .pipeline.pipeline import Pipeline, ACTOR

from .service.extra import BeforeHandler, AfterHandler
from .service.group import ServiceGroup
from .service.service import Service, to_service



"""
Script
------
The Script module provides a set of `pydantic` models for representing the dialog graph.
These models are used to define the conversation flow, and to determine the appropriate response based on
the user's input and the current state of the conversation.
"""

# %%
from __future__ import annotations
from enum import Enum
import inspect
import logging
from typing import Callable, List, Optional, Any, Dict, Tuple, Union, TYPE_CHECKING

from pydantic import BaseModel, field_validator, validate_call, Field

from .types import Label, LabelType, ConditionType, ConstLabel  # noqa: F401
from .message import Message
from .keywords import Keywords
from .normalization import normalize_condition, normalize_label

if TYPE_CHECKING:
    from chatsky.script.core.context import Context
    from chatsky.pipeline.pipeline.pipeline import Pipeline

logger = logging.getLogger(__name__)


class UserFunctionType(str, Enum):
    LABEL = "label"
    RESPONSE = "response"
    CONDITION = "condition"
    TRANSITION_PROCESSING = "pre_transitions_processing"
    RESPONSE_PROCESSING = "pre_response_processing"


USER_FUNCTION_TYPES: Dict[UserFunctionType, Tuple[Tuple[str, ...], str]] = {
    UserFunctionType.LABEL: (("Context", "Pipeline"), "ConstLabel"),
    UserFunctionType.RESPONSE: (("Context", "Pipeline"), "Message"),
    UserFunctionType.CONDITION: (("Context", "Pipeline"), "bool"),
    UserFunctionType.RESPONSE_PROCESSING: (("Context", "Pipeline"), "None"),
    UserFunctionType.TRANSITION_PROCESSING: (("Context", "Pipeline"), "None"),
}


def _types_equal(signature_type: Any, expected_type: str) -> bool:
    """
    This function checks equality of signature type with expected type.
    Three cases are handled. If no signature is present, it is presumed that types are equal.
    If signature is a type, it is compared with expected type as is.
    If signature is a string, it is compared with expected type name.

    :param signature_type: type received from function signature.
    :param expected_type: expected type - a class.
    :return: true if types are equal, false otherwise.
    """
    signature_str = signature_type.__name__ if hasattr(signature_type, "__name__") else str(signature_type)
    signature_empty = signature_type == inspect.Parameter.empty
    expected_string = signature_str == expected_type
    expected_global = str(signature_type) == str(globals().get(expected_type))
    return signature_empty or expected_string or expected_global


def _validate_callable(callable: Callable, func_type: UserFunctionType, flow_label: str, node_label: str) -> List:
    """
    This function validates a function during :py:class:`~chatsky.script.Script` validation.
    It checks parameter number (unconditionally), parameter types (if specified) and return type (if specified).

    :param callable: Function to validate.
    :param func_type: Type of the function (label, condition, response, etc.).
    :param flow_label: Flow label this function is related to (used for error localization only).
    :param node_label: Node label this function is related to (used for error localization only).
    :return: list of produced error messages.
    """

    error_msgs = list()
    signature = inspect.signature(callable)
    arguments_type, return_type = USER_FUNCTION_TYPES[func_type]
    params = list(signature.parameters.values())
    if len(params) != len(arguments_type):
        msg = (
            f"Incorrect parameter number for {callable.__name__!r}: "
            f"should be {len(arguments_type)}, not {len(params)}. "
            f"Error found at {(flow_label, node_label)!r}."
        )
        error_msgs.append(msg)
    for idx, param in enumerate(params):
        if not _types_equal(param.annotation, arguments_type[idx]):
            msg = (
                f"Incorrect parameter annotation for parameter #{idx + 1} "
                f" of {callable.__name__!r}: "
                f"should be {arguments_type[idx]}, not {param.annotation}. "
                f"Error found at {(flow_label, node_label)!r}."
            )
            error_msgs.append(msg)
    if not _types_equal(signature.return_annotation, return_type):
        msg = (
            f"Incorrect return type annotation of {callable.__name__!r}: "
            f"should be {return_type!r}, not {signature.return_annotation}. "
            f"Error found at {(flow_label, node_label)!r}."
        )
        error_msgs.append(msg)
    return error_msgs


class Node(BaseModel, extra="forbid", validate_assignment=True):
    """
    The class for the `Node` object.
    """

    transitions: Dict[Label, ConditionType] = Field(default_factory=dict, alias="TRANSITIONS")
    response: Union[Message, Callable[[Context, Pipeline], Message]] = Field(default_factory=Message, alias="RESPONSE")
    pre_transitions_processing: Dict[Any, Callable] = Field(default_factory=dict, alias="PRE_TRANSITIONS_PROCESSING")
    pre_response_processing: Dict[Any, Callable] = Field(default_factory=dict, alias="PRE_RESPONSE_PROCESSING")
    misc: dict = Field(default_factory=dict, alias="MISC")

    @field_validator("transitions", mode="before")
    @classmethod
    @validate_call
    def normalize_transitions(cls, transitions: Dict[Label, ConditionType]) -> Dict[Label, Callable]:
        """
        The function which is used to normalize transitions and returns normalized dict.

        :param transitions: Transitions to normalize.
        :return: Transitions with normalized label and condition.
        """
        transitions = {
            normalize_label(label): normalize_condition(condition) for label, condition in transitions.items()
        }
        return transitions


class Script(BaseModel, extra="forbid"):
    """
    The class for the `Script` object.
    """

    script: Dict[LabelType, Dict[LabelType, Node]]

    @field_validator("script", mode="before")
    @classmethod
    @validate_call
    def normalize_script(cls, script: Dict[LabelType, Any]) -> Dict[LabelType, Dict[LabelType, Dict[str, Any]]]:
        """
        This function normalizes :py:class:`.Script`: it returns dict where the GLOBAL node is moved
        into the flow with the GLOBAL name. The function returns the structure

        `{GLOBAL: {...NODE...}, ...}` -> `{GLOBAL: {GLOBAL: {...NODE...}}, ...}`.

        :param script: :py:class:`.Script` that describes the dialog scenario.
        :return: Normalized :py:class:`.Script`.
        """
        if isinstance(script, dict):
            if Keywords.GLOBAL in script and all(
                [isinstance(item, Keywords) for item in script[Keywords.GLOBAL].keys()]
            ):
                script[Keywords.GLOBAL] = {Keywords.GLOBAL: script[Keywords.GLOBAL]}
        return script

    @field_validator("script", mode="before")
    @classmethod
    @validate_call
    def validate_script_before(cls, script: Dict[LabelType, Any]) -> Dict[LabelType, Dict[LabelType, Dict[str, Any]]]:
        error_msgs = []
        for flow_name, flow in script.items():
            for node_name, node in flow.items():
                # validate labeling
                transitions = node.get("transitions", dict())
                for label in transitions.keys():
                    if callable(label):
                        error_msgs += _validate_callable(label, UserFunctionType.LABEL, flow_name, node_name)

                # validate responses
                response = node.get("response", None)
                if callable(response):
                    error_msgs += _validate_callable(
                        response,
                        UserFunctionType.RESPONSE,
                        flow_name,
                        node_name,
                    )

                # validate conditions
                for label, condition in transitions.items():
                    if callable(condition):
                        error_msgs += _validate_callable(
                            condition,
                            UserFunctionType.CONDITION,
                            flow_name,
                            node_name,
                        )

                # validate pre_transitions- and pre_response_processing
                pre_transitions_processing = node.get("pre_transitions_processing", dict())
                pre_response_processing = node.get("pre_response_processing", dict())
                for place, functions in zip(
                    (UserFunctionType.TRANSITION_PROCESSING, UserFunctionType.RESPONSE_PROCESSING),
                    (pre_transitions_processing, pre_response_processing),
                ):
                    for function in functions.values():
                        if callable(function):
                            error_msgs += _validate_callable(
                                function,
                                place,
                                flow_name,
                                node_name,
                            )
        if error_msgs:
            error_number_string = "1 error" if len(error_msgs) == 1 else f"{len(error_msgs)} errors"
            raise ValueError(
                f"Found {error_number_string}:\n" + "\n".join([f"{i}) {er}" for i, er in enumerate(error_msgs, 1)])
            )
        else:
            return script

    @field_validator("script", mode="after")
    @classmethod
    @validate_call
    def validate_script_after(cls, script: Dict[LabelType, Any]) -> Dict[LabelType, Dict[LabelType, Dict[str, Any]]]:
        error_msgs = []
        for flow_name, flow in script.items():
            for node_name, node in flow.items():
                # validate labeling
                for label in node.transitions.keys():
                    if not callable(label):
                        norm_flow_label, norm_node_label, _ = normalize_label(label, flow_name)
                        if norm_flow_label not in script.keys():
                            msg = (
                                f"Flow {norm_flow_label!r} cannot be found for label={label}. "
                                f"Error found at {(flow_name, node_name)!r}."
                            )
                        elif norm_node_label not in script[norm_flow_label].keys():
                            msg = (
                                f"Node {norm_node_label!r} cannot be found for label={label}. "
                                f"Error found at {(flow_name, node_name)!r}."
                            )
                        else:
                            msg = None
                        if msg is not None:
                            error_msgs.append(msg)

        if error_msgs:
            error_number_string = "1 error" if len(error_msgs) == 1 else f"{len(error_msgs)} errors"
            raise ValueError(
                f"Found {error_number_string}:\n" + "\n".join([f"{i}) {er}" for i, er in enumerate(error_msgs, 1)])
            )
        else:
            return script

    def __getitem__(self, key):
        return self.script[key]

    def get(self, key, value=None):
        return self.script.get(key, value)

    def keys(self):
        return self.script.keys()

    def items(self):
        return self.script.items()

    def values(self):
        return self.script.values()

    def __iter__(self):
        return self.script.__iter__()



"""
Labels
------
:py:const:`Labels <chatsky.script.ConstLabel>` are one of the important components of the dialog graph,
which determine the targeted node name of the transition.
They are used to identify the next step in the conversation.
Labels can also be used in combination with other conditions,
such as the current context or user data, to create more complex and dynamic conversations.

This module contains a standard set of scripting :py:const:`labels <chatsky.script.ConstLabel>` that
can be used by developers to define the conversation flow.
"""

from __future__ import annotations
from typing import Optional, Callable, TYPE_CHECKING
from chatsky.script import Context, ConstLabel

if TYPE_CHECKING:
    from chatsky.pipeline.pipeline.pipeline import Pipeline


def repeat(priority: Optional[float] = None) -> Callable[[Context, Pipeline], ConstLabel]:
    """
    Returns transition handler that takes :py:class:`.Context`,
    :py:class:`~chatsky.pipeline.Pipeline` and :py:const:`priority <float>`.
    This handler returns a :py:const:`label <ConstLabel>`
    to the last node with a given :py:const:`priority <float>`.
    If the priority is not given, `Pipeline.actor.label_priority` is used as default.

    :param priority: Priority of transition. Uses `Pipeline.actor.label_priority` if priority not defined.
    """

    def repeat_transition_handler(ctx: Context, pipeline: Pipeline) -> ConstLabel:
        current_priority = pipeline.actor.label_priority if priority is None else priority
        if len(ctx.labels) >= 1:
            flow_label, label = list(ctx.labels.values())[-1]
        else:
            flow_label, label = pipeline.actor.start_label[:2]
        return (flow_label, label, current_priority)

    return repeat_transition_handler


def previous(priority: Optional[float] = None) -> Callable[[Context, Pipeline], ConstLabel]:
    """
    Returns transition handler that takes :py:class:`~chatsky.script.Context`,
    :py:class:`~chatsky.pipeline.Pipeline` and :py:const:`priority <float>`.
    This handler returns a :py:const:`label <chatsky.script.ConstLabel>`
    to the previous node with a given :py:const:`priority <float>`.
    If the priority is not given, `Pipeline.actor.label_priority` is used as default.
    If the current node is the start node, fallback is returned.

    :param priority: Priority of transition. Uses `Pipeline.actor.label_priority` if priority not defined.
    """

    def previous_transition_handler(ctx: Context, pipeline: Pipeline) -> ConstLabel:
        current_priority = pipeline.actor.label_priority if priority is None else priority
        if len(ctx.labels) >= 2:
            flow_label, label = list(ctx.labels.values())[-2]
        elif len(ctx.labels) == 1:
            flow_label, label = pipeline.actor.start_label[:2]
        else:
            flow_label, label = pipeline.actor.fallback_label[:2]
        return (flow_label, label, current_priority)

    return previous_transition_handler


def to_start(priority: Optional[float] = None) -> Callable[[Context, Pipeline], ConstLabel]:
    """
    Returns transition handler that takes :py:class:`~chatsky.script.Context`,
    :py:class:`~chatsky.pipeline.Pipeline` and :py:const:`priority <float>`.
    This handler returns a :py:const:`label <chatsky.script.ConstLabel>`
    to the start node with a given :py:const:`priority <float>`.
    If the priority is not given, `Pipeline.actor.label_priority` is used as default.

    :param priority: Priority of transition. Uses `Pipeline.actor.label_priority` if priority not defined.
    """

    def to_start_transition_handler(ctx: Context, pipeline: Pipeline) -> ConstLabel:
        current_priority = pipeline.actor.label_priority if priority is None else priority
        return (*pipeline.actor.start_label[:2], current_priority)

    return to_start_transition_handler


def to_fallback(priority: Optional[float] = None) -> Callable[[Context, Pipeline], ConstLabel]:
    """
    Returns transition handler that takes :py:class:`~chatsky.script.Context`,
    :py:class:`~chatsky.pipeline.Pipeline` and :py:const:`priority <float>`.
    This handler returns a :py:const:`label <chatsky.script.ConstLabel>`
    to the fallback node with a given :py:const:`priority <float>`.
    If the priority is not given, `Pipeline.actor.label_priority` is used as default.

    :param priority: Priority of transition. Uses `Pipeline.actor.label_priority` if priority not defined.
    """

    def to_fallback_transition_handler(ctx: Context, pipeline: Pipeline) -> ConstLabel:
        current_priority = pipeline.actor.label_priority if priority is None else priority
        return (*pipeline.actor.fallback_label[:2], current_priority)

    return to_fallback_transition_handler


def _get_label_by_index_shifting(
    ctx: Context,
    pipeline: Pipeline,
    priority: Optional[float] = None,
    increment_flag: bool = True,
    cyclicality_flag: bool = True,
) -> ConstLabel:
    """
    Function that returns node label from the context and pipeline after shifting the index.

    :param ctx: Dialog context.
    :param pipeline: Dialog pipeline.
    :param priority: Priority of transition. Uses `Pipeline.actor.label_priority` if priority not defined.
    :param increment_flag: If it is `True`, label index is incremented by `1`,
        otherwise it is decreased by `1`. Defaults to `True`.
    :param cyclicality_flag: If it is `True` the iteration over the label list is going cyclically
        (e.g the element with `index = len(labels)` has `index = 0`). Defaults to `True`.
    :return: The tuple that consists of `(flow_label, label, priority)`.
        If fallback is executed `(flow_fallback_label, fallback_label, priority)` are returned.
    """
    flow_label, node_label, current_priority = repeat(priority)(ctx, pipeline)
    labels = list(pipeline.script.get(flow_label, {}))

    if node_label not in labels:
        return (*pipeline.actor.fallback_label[:2], current_priority)

    label_index = labels.index(node_label)
    label_index = label_index + 1 if increment_flag else label_index - 1
    if not (cyclicality_flag or (0 <= label_index < len(labels))):
        return (*pipeline.actor.fallback_label[:2], current_priority)
    label_index %= len(labels)

    return (flow_label, labels[label_index], current_priority)


def forward(
    priority: Optional[float] = None, cyclicality_flag: bool = True
) -> Callable[[Context, Pipeline], ConstLabel]:
    """
    Returns transition handler that takes :py:class:`~chatsky.script.Context`,
    :py:class:`~chatsky.pipeline.Pipeline` and :py:const:`priority <float>`.
    This handler returns a :py:const:`label <chatsky.script.ConstLabel>`
    to the forward node with a given :py:const:`priority <float>` and :py:const:`cyclicality_flag <bool>`.
    If the priority is not given, `Pipeline.actor.label_priority` is used as default.

    :param priority: Float priority of transition. Uses `Pipeline.actor.label_priority` if priority not defined.
    :param cyclicality_flag: If it is `True`, the iteration over the label list is going cyclically
        (e.g the element with `index = len(labels)` has `index = 0`). Defaults to `True`.
    """

    def forward_transition_handler(ctx: Context, pipeline: Pipeline) -> ConstLabel:
        return _get_label_by_index_shifting(
            ctx, pipeline, priority, increment_flag=True, cyclicality_flag=cyclicality_flag
        )

    return forward_transition_handler


def backward(
    priority: Optional[float] = None, cyclicality_flag: bool = True
) -> Callable[[Context, Pipeline], ConstLabel]:
    """
    Returns transition handler that takes :py:class:`~chatsky.script.Context`,
    :py:class:`~chatsky.pipeline.Pipeline` and :py:const:`priority <float>`.
    This handler returns a :py:const:`label <chatsky.script.ConstLabel>`
    to the backward node with a given :py:const:`priority <float>` and :py:const:`cyclicality_flag <bool>`.
    If the priority is not given, `Pipeline.actor.label_priority` is used as default.

    :param priority: Float priority of transition. Uses `Pipeline.actor.label_priority` if priority not defined.
    :param cyclicality_flag: If it is `True`, the iteration over the label list is going cyclically
        (e.g the element with `index = len(labels)` has `index = 0`). Defaults to `True`.
    """

    def back_transition_handler(ctx: Context, pipeline: Pipeline) -> ConstLabel:
        return _get_label_by_index_shifting(
            ctx, pipeline, priority, increment_flag=False, cyclicality_flag=cyclicality_flag
        )

    return back_transition_handler



"""
Utils
-----
This module includes utility functions designed for statistics collection.

The functions below can be used to configure the opentelemetry destination.

.. code:: python

    set_logger_destination(OTLPLogExporter("grpc://localhost:4317", insecure=True))
    set_tracer_destination(OTLPSpanExporter("grpc://localhost:4317", insecure=True))

"""

import json
import getpass
from urllib import parse
from typing import Optional, Tuple
from argparse import Namespace, Action

import requests
from opentelemetry.sdk.resources import Resource
from opentelemetry._logs import get_logger_provider, set_logger_provider
from opentelemetry.trace import get_tracer_provider, set_tracer_provider
from opentelemetry.metrics import get_meter_provider, set_meter_provider
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk._logs import LoggerProvider
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter, SpanExporter
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter, LogExporter
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter, MetricExporter

from chatsky.pipeline import ExtraHandlerRuntimeInfo

SERVICE_NAME = "chatsky"

resource = Resource.create({"service.name": SERVICE_NAME})
"""
Singletone :py:class:`~Resource` instance shared inside the framework.
"""
tracer_provider = TracerProvider(resource=resource)
"""
Global tracer provider bound to the Chatsky resource.
"""
logger_provider = LoggerProvider(resource=resource)
"""
Global logger provider bound to the Chatsky resource.
"""
set_logger_provider(logger_provider)
set_tracer_provider(tracer_provider)


def set_logger_destination(exporter: Optional[LogExporter] = None):
    """
    Configure the global Opentelemetry logger provider to export logs to the given destination.

    :param exporter: Opentelemetry log exporter instance.
    """
    if exporter is None:
        exporter = OTLPLogExporter(endpoint="grpc://localhost:4317", insecure=True)
    get_logger_provider().add_log_record_processor(BatchLogRecordProcessor(exporter))


def set_meter_destination(exporter: Optional[MetricExporter] = None):
    """
    Configure the global Opentelemetry meter provider to export metrics to the given destination.

    :param exporter: Opentelemetry meter exporter instance.
    """
    if exporter is None:
        exporter = OTLPMetricExporter(endpoint="grpc://localhost:4317", insecure=True)
    cur_meter_provider = get_meter_provider()
    new_meter_provider = MeterProvider(resource=resource, metric_readers=[PeriodicExportingMetricReader(exporter)])
    if not isinstance(cur_meter_provider, MeterProvider):
        set_meter_provider(new_meter_provider)


def set_tracer_destination(exporter: Optional[SpanExporter] = None):
    """
    Configure the global Opentelemetry tracer provider to export traces to the given destination.

    :param exporter: Opentelemetry span exporter instance.
    """
    if exporter is None:
        exporter = OTLPSpanExporter(endpoint="grpc://localhost:4317", insecure=True)
    get_tracer_provider().add_span_processor(BatchSpanProcessor(exporter))


def get_extra_handler_name(info: ExtraHandlerRuntimeInfo, postfix: str = "") -> str:
    """
    This function can be used to obtain a key, under which the wrapper data will be stored
    in the context.

    :param info: Handler runtime info obtained from the pipeline.
    :param postfix: Field-specific postfix that will be appended to the field name.
    """

    path = info.component.path.replace(".", "-")
    return f"{path}" + (f"-{postfix}" if postfix else "")


def get_superset_session(args: Namespace, base_url: str = "http://localhost:8088/") -> Tuple[requests.Session, dict]:
    """
    Utility function for authorized interaction with Superset HTTP API.

    :param args: Command line arguments including Superset username and Superset password.
    :param base_url: Base Superset URL.

    :return: Authorized session - authorization headers tuple.
    """
    healthcheck_url = parse.urljoin(base_url, "/healthcheck")
    login_url = parse.urljoin(base_url, "/api/v1/security/login")
    csrf_url = parse.urljoin(base_url, "/api/v1/security/csrf_token/")

    session = requests.Session()
    # do healthcheck
    response = session.get(healthcheck_url, timeout=10)
    response.raise_for_status()
    # get access token
    access_request = session.post(
        login_url,
        headers={"Content-Type": "application/json", "Accept": "*/*"},
        data=json.dumps({"username": args.username, "password": args.password, "refresh": True, "provider": "db"}),
    )
    access_token = access_request.json()["access_token"]
    # get csrf_token
    csrf_request = session.get(csrf_url, headers={"Authorization": f"Bearer {access_token}"})
    csrf_token = csrf_request.json()["result"]
    headers = {
        "Authorization": f"Bearer {access_token}",
        "X-CSRFToken": csrf_token,
    }
    return session, headers


def drop_superset_assets(session: requests.Session, headers: dict, base_url: str):
    """
    Drop the existing assets from the Superset dashboard.

    :param session: Authorized Superset session.
    :param headers: Superset session headers.
    :param base_url: Base Superset URL.
    """
    dashboard_url = parse.urljoin(base_url, "/api/v1/dashboard")
    charts_url = parse.urljoin(base_url, "/api/v1/chart")
    datasets_url = parse.urljoin(base_url, "/api/v1/dataset")
    database_url = parse.urljoin(base_url, "/api/v1/database/")
    delete_res: requests.Response

    dashboard_res = session.get(dashboard_url, headers=headers)
    dashboard_json = dashboard_res.json()
    if dashboard_json["count"] > 0:
        delete_res = requests.delete(dashboard_url, params={"q": json.dumps(dashboard_json["ids"])}, headers=headers)
        delete_res.raise_for_status()

    charts_result = session.get(charts_url, headers=headers)
    charts_json = charts_result.json()
    if charts_json["count"] > 0:
        delete_res = requests.delete(charts_url, params={"q": json.dumps(charts_json["ids"])}, headers=headers)
        delete_res.raise_for_status()

    datasets_result = session.get(datasets_url, headers=headers)
    datasets_json = datasets_result.json()
    if datasets_json["count"] > 0:
        delete_res = requests.delete(datasets_url, params={"q": json.dumps(datasets_json["ids"])}, headers=headers)
        delete_res.raise_for_status()

    database_res = session.get(database_url, headers=headers)
    database_json = database_res.json()
    if database_json["count"] > 0:
        delete_res = requests.delete(database_url + str(database_json["ids"][-1]), headers=headers)
        delete_res.raise_for_status()


class PasswordAction(Action):
    """
    Child class for Argparse's :py:class:`~Action` that prompts users for passwords interactively,
    ensuring password safety, unless the password is specified directly.

    """

    def __init__(
        self, option_strings, dest=None, nargs=0, default=None, required=False, type=None, metavar=None, help=None
    ):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            default=default,
            required=required,
            metavar=metavar,
            type=type,
            help=help,
        )

    def __call__(self, parser, args, values, option_string=None):
        if values:
            print(f"{self.dest}: setting passwords explicitly through the command line is discouraged.")
            setattr(args, self.dest, values)
        else:
            setattr(args, self.dest, getpass.getpass(prompt=f"{self.dest}: "))



"""
Processing
---------------------------
This module provides wrappers for :py:class:`~chatsky.slots.slots.SlotManager`'s API.
"""

from __future__ import annotations

import logging
from typing import Awaitable, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from chatsky.slots.slots import SlotName
    from chatsky.script import Context
    from chatsky.pipeline import Pipeline

logger = logging.getLogger(__name__)


def extract(*slots: SlotName) -> Callable[[Context, Pipeline], Awaitable[None]]:
    """
    Extract slots listed slots.
    This will override all slots even if they are already extracted.

    :param slots: List of slot names to extract.
    """

    async def inner(ctx: Context, pipeline: Pipeline) -> None:
        manager = ctx.framework_data.slot_manager
        for slot in slots:  # todo: maybe gather
            await manager.extract_slot(slot, ctx, pipeline)

    return inner


def extract_all():
    """
    Extract all slots defined in the pipeline.
    """

    async def inner(ctx: Context, pipeline: Pipeline):
        manager = ctx.framework_data.slot_manager
        await manager.extract_all(ctx, pipeline)

    return inner


def unset(*slots: SlotName) -> Callable[[Context, Pipeline], None]:
    """
    Mark specified slots as not extracted and clear extracted values.

    :param slots: List of slot names to extract.
    """

    def unset_inner(ctx: Context, pipeline: Pipeline) -> None:
        manager = ctx.framework_data.slot_manager
        for slot in slots:
            manager.unset_slot(slot)

    return unset_inner


def unset_all():
    """
    Mark all slots as not extracted and clear all extracted values.
    """

    def inner(ctx: Context, pipeline: Pipeline):
        manager = ctx.framework_data.slot_manager
        manager.unset_all_slots()

    return inner


def fill_template() -> Callable[[Context, Pipeline], None]:
    """
    Fill the response template in the current node.

    Response message of the current node should be a format-string: e.g. "Your username is {profile.username}".
    """

    def inner(ctx: Context, pipeline: Pipeline) -> None:
        manager = ctx.framework_data.slot_manager
        # get current node response
        response = ctx.current_node.response

        if response is None:
            return

        if callable(response):
            response = response(ctx, pipeline)

        new_text = manager.fill_template(response.text)

        response.text = new_text
        ctx.current_node.response = response

    return inner



"""
Utils
-----
The Utils module contains several service functions that are commonly used throughout the framework.
These functions provide a variety of utility functionality.
"""

import collections
from typing import Union, List
from inspect import isfunction

from ..service.service import Service
from ..service.group import ServiceGroup


def pretty_format_component_info_dict(
    service: dict,
    show_extra_handlers: bool,
    offset: str = "",
    extra_handlers_key: str = "extra_handlers",
    type_key: str = "type",
    name_key: str = "name",
    indent: int = 4,
) -> str:
    """
    Function for dumping any pipeline components info dictionary (received from `info_dict` property) as a string.
    Resulting string is formatted with YAML-like format, however it's not strict and shouldn't be parsed.
    However, most preferable usage is via `pipeline.pretty_format`.

    :param service: (required) Pipeline components info dictionary.
    :param show_extra_handlers: (required) Whether to include Extra Handlers or not (could be many and/or generated).
    :param offset: Current level new line offset.
    :param extra_handlers_key: Key that is mapped to Extra Handlers lists.
    :param type_key: Key that is mapped to components type name.
    :param name_key: Key that is mapped to components name.
    :param indent: Current level new line offset (whitespace number).
    :return: Formatted string
    """
    indent = " " * indent
    representation = f"{offset}{service.get(type_key, '[None]')}%s:\n" % (
        f" '{service.get(name_key, '[None]')}'" if name_key in service else ""
    )
    for key, value in service.items():
        if key not in (type_key, name_key, extra_handlers_key) or (key == extra_handlers_key and show_extra_handlers):
            if isinstance(value, List):
                if len(value) > 0:
                    values = [
                        pretty_format_component_info_dict(instance, show_extra_handlers, f"{indent * 2}{offset}")
                        for instance in value
                    ]
                    value_str = "\n%s" % "\n".join(values)
                else:
                    value_str = "[None]"
            else:
                value_str = str(value)
            representation += f"{offset}{indent}{key}: {value_str}\n"
    return representation[:-1]


def rename_component_incrementing(
    service: Union[Service, ServiceGroup], collisions: List[Union[Service, ServiceGroup]]
) -> str:
    """
    Function for generating new name for a pipeline component,
    that has similar name with other components in the same group.
    The name is generated according to these rules:

    - If service's handler is "ACTOR", it is named `actor`.
    - If service's handler is `Callable`, it is named after this `callable`.
    - If it's a service group, it is named `service_group`.
    - Otherwise, it is names `noname_service`.
    - | After that, `_[NUMBER]` is added to the resulting name,
        where `_[NUMBER]` is number of components with the same name in current service group.

    :param service: Service to be renamed.
    :param collisions: Services in the same service group as service.
    :return: Generated name
    """
    if isinstance(service, Service) and isinstance(service.handler, str) and service.handler == "ACTOR":
        base_name = "actor"
    elif isinstance(service, Service) and callable(service.handler):
        if isfunction(service.handler):
            base_name = service.handler.__name__
        else:
            base_name = service.handler.__class__.__name__
    elif isinstance(service, ServiceGroup):
        base_name = "service_group"
    else:
        base_name = "noname_service"

    name_index = 0
    while f"{base_name}_{name_index}" in [component.name for component in collisions]:
        name_index += 1
    return f"{base_name}_{name_index}"


def finalize_service_group(service_group: ServiceGroup, path: str = ".") -> bool:
    """
    Function that iterates through a service group (and all its subgroups),
    finalizing component's names and paths in it.
    Components are renamed only if user didn't set a name for them. Their paths are also generated here.
    It also searches for "ACTOR" in the group, throwing exception if no actor or multiple actors found.

    :param service_group: Service group to resolve name collisions in.
    :param path:
        A prefix for component paths -- path of `component` is equal to `{path}.{component.name}`.
        Defaults to ".".
    """
    actor = False
    names_counter = collections.Counter([component.name for component in service_group.components])
    for component in service_group.components:
        if component.name is None:
            component.name = rename_component_incrementing(component, service_group.components)
        elif names_counter[component.name] > 1:
            raise Exception(f"User defined service name collision ({path})!")
        component.path = f"{path}.{component.name}"

        if isinstance(component, Service) and isinstance(component.handler, str) and component.handler == "ACTOR":
            actor_found = True
        elif isinstance(component, ServiceGroup):
            actor_found = finalize_service_group(component, f"{path}.{component.name}")
        else:
            actor_found = False

        if actor_found:
            if not actor:
                actor = actor_found
            else:
                raise Exception(f"More than one actor found in group ({path})!")
    return actor



"""
Keywords
--------
Keywords are used to define the dialog graph, which is the structure of a conversation.
They are used to determine all nodes in the script and to assign python objects and python functions for nodes.

"""

from enum import Enum


class Keywords(str, Enum):
    """
    Keywords used to define the dialog script (:py:class:`~chatsky.script.Script`).
    The data type `dict` is used to describe the scenario.
    `Enums` of this class are used as keys in this `dict`.
    Different keys correspond to the different value types aimed at different purposes.

    Enums:

    GLOBAL: Enum(auto)
        This keyword is used to define a global node.
        The value that corresponds to this key has the `dict` type with keywords:

        `{TRANSITIONS:..., RESPONSE:..., PRE_RESPONSE_PROCESSING:..., MISC:...}`.
        There can be only one global node in a script :py:class:`~chatsky.script.Script`.
        The global node is defined at the flow level as opposed to regular nodes.
        This node allows to define default global values for all nodes.

    LOCAL: Enum(auto)
        This keyword is used to define the local node.
        The value that corresponds to this key has the `dict` type with keywords:

        `{TRANSITIONS:..., RESPONSE:..., PRE_RESPONSE_PROCESSING:..., MISC:...}`.
        The local node is defined in the same way as all other nodes in the flow of this node.
        It also allows to redefine default values for all nodes in this node's flow.

    TRANSITIONS: Enum(auto)
        This keyword defines possible transitions from node.
        The value that corresponds to the `TRANSITIONS` key has the `dict` type.
        Every key-value pair describes the transition node and the condition:

        `{label_to_transition_0: condition_for_transition_0, ..., label_to_transition_N: condition_for_transition_N}`,

        where `label_to_transition_i` is a node into which the actor make the transition in case of
        `condition_for_transition_i == True`.

    RESPONSE: Enum(auto)
        The keyword specifying the result which is returned to the user after getting to the node.
        Value corresponding to the `RESPONSE` key can have any data type.

    MISC: Enum(auto)
        The keyword specifying `dict` containing extra data,
        which were not aimed to be used in the standard functions of `DFE`.
        Value corresponding to the `MISC` key must have `dict` type:

        `{"VAR_KEY_0": VAR_VALUE_0, ..., "VAR_KEY_N": VAR_VALUE_N}`,

        where `"VAR_KEY_0"` is an arbitrary name of the value which is saved into the `MISC`.

    PRE_RESPONSE_PROCESSING: Enum(auto)
        The keyword specifying the preprocessing that is called before the response generation.
        The value that corresponds to the `PRE_RESPONSE_PROCESSING` key must have the `dict` type:

        `{"PRE_RESPONSE_PROC_0": pre_response_proc_func_0, ..., "PRE_RESPONSE_PROC_N": pre_response_proc__func_N}`,

        where `"PRE_RESPONSE_PROC_i"` is an arbitrary name of the preprocessing stage in the pipeline.
        Unless the :py:class:`~chatsky.pipeline.pipeline.Pipeline`'s `parallelize_processing` flag
        is set to `True`, calls to `pre_response_proc__func_i` are made in-order.

    PRE_TRANSITIONS_PROCESSING: Enum(auto)
        The keyword specifying the preprocessing that is called before the transition.
        The value that corresponds to the `PRE_TRANSITIONS_PROCESSING` key must have the `dict` type:

        `{"PRE_TRANSITIONS_PROC_0": pre_transitions_proc_func_0, ...,
        "PRE_TRANSITIONS_PROC_N": pre_transitions_proc_func_N}`,

        where `"PRE_TRANSITIONS_PROC_i"` is an arbitrary name of the preprocessing stage in the pipeline.
        Unless the :py:class:`~chatsky.pipeline.pipeline.Pipeline`'s `parallelize_processing` flag
        is set to `True`, calls to `pre_transitions_proc_func_i` are made in-order.

    """

    GLOBAL = "GLOBAL"
    LOCAL = "LOCAL"
    TRANSITIONS = "TRANSITIONS"
    RESPONSE = "RESPONSE"
    MISC = "MISC"
    PRE_RESPONSE_PROCESSING = "PRE_RESPONSE_PROCESSING"
    PRE_TRANSITIONS_PROCESSING = "PRE_TRANSITIONS_PROCESSING"


# Redefine shortcuts
GLOBAL = Keywords.GLOBAL
LOCAL = Keywords.LOCAL
TRANSITIONS = Keywords.TRANSITIONS
RESPONSE = Keywords.RESPONSE
MISC = Keywords.MISC
PRE_RESPONSE_PROCESSING = Keywords.PRE_RESPONSE_PROCESSING
PRE_TRANSITIONS_PROCESSING = Keywords.PRE_TRANSITIONS_PROCESSING



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

#
# This is an example "local" configuration file. In order to set/override config
# options that ONLY apply to your local environment, simply copy/rename this file
# to docker/pythonpath/superset_config_docker.py
# It ends up being imported by docker/superset_config.py which is loaded by
# superset/config.py
#
import os

SQLALCHEMY_DATABASE_URI = "postgresql+psycopg2://{0}:{1}@dashboard-metadata:{2}/{3}".format(
    os.getenv("POSTGRES_USERNAME"),
    os.getenv("POSTGRES_PASSWORD"),
    os.getenv("SUPERSET_METADATA_PORT"),
    os.getenv("POSTGRES_DB"),
)
